"""Module with the classes that manage the authorStats
and paperStats plots.

This file is part of the physbiblio package.
"""
import traceback
import numpy as np
import matplotlib
matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_qt5agg import \
	FigureCanvasQTAgg as FigureCanvas
from PySide2.QtCore import Qt
from PySide2.QtGui import QFont
from PySide2.QtWidgets import QDialog, QGridLayout, QLineEdit, QPushButton

try:
	from physbiblio.database import pBDB
	from physbiblio.config import pbConfig
	from physbiblio.inspireStats import pBStats
	from physbiblio.gui.basicDialogs import askDirName, infoMessage
	from physbiblio.gui.commonClasses import PBLabel
except ImportError:
	print("Could not find physbiblio and its modules!")
	print(traceback.format_exc())

figTitles = [
	"Paper number",
	"Papers per year",
	"Total citations",
	"Citations per year",
	"Mean citations",
	"Citations for each paper"
]


class AuthorStatsPlots(QDialog):
	"""Class that constructs a window to show
	the results of `authorStats`
	"""

	def __init__(self, figs, title=None, parent=None):
		"""Constructor.

		Parameters:
			figs: the list of figures
			title (optional): the window title
			parent (default None): the parent object,
				which should have a property `lastAuthorStats`
		"""
		super(AuthorStatsPlots, self).__init__(parent)
		self.figs = figs
		if title is not None:
			self.setWindowTitle(title)
		layout = QGridLayout(self)
		layout.setSpacing(1)
		self.setLayout(layout)
		self.updatePlots()

		nlines = int(len(figs)/2)
		self.layout().addWidget(PBLabel(
			"Click on the lines to have more information:"), nlines + 1, 0)

		try:
			hIndex = "%d"%self.parent().lastAuthorStats["h"]
		except (TypeError, AttributeError):
			hIndex = "ND"
		self.hIndex = PBLabel("Author h index: %s"%hIndex)
		largerFont = QFont("Times", 15, QFont.Bold)
		self.hIndex.setFont(largerFont)
		self.layout().addWidget(self.hIndex, nlines + 1, 1)

		self.textBox = QLineEdit("")
		self.textBox.setReadOnly(True)
		self.layout().addWidget(self.textBox, nlines + 2, 0, 1, 2)
		self.saveButton = QPushButton('Save', self)
		self.saveButton.clicked.connect(self.saveAction)
		self.layout().addWidget(self.saveButton, nlines + 3, 0)

		self.clButton = QPushButton('Close', self)
		self.clButton.clicked.connect(self.onClose)
		self.clButton.setAutoDefault(True)
		self.layout().addWidget(self.clButton, nlines + 3, 1)

	def onClose(self):
		"""Close dialog"""
		QDialog.close(self)

	def saveAction(self):
		"""Save the plot into a file,
		after asking the directory where to save them
		"""
		savePath = askDirName(self,
			"Where do you want to save the plots of the stats?")
		if savePath != "":
			try:
				self.parent().lastAuthorStats["figs"] = pBStats.plotStats(
					author=True, save=True, path=savePath)
			except AttributeError:
				pBLogger.warning("", exc_info=True)
			infoMessage("Plots saved.")
			self.saveButton.setDisabled(True)

	def onPress(self, event):
		"""Print the plot coordinates where a click was performed.
		To be connected through `mpl_connect`.
		Used for testing.

		Parameter:
			a `matplotlib.backend_bases.Event`
		"""
		print(event.xdata, event.ydata)

	def pickEvent(self, event):
		"""Save into `self.textBox` the coordinates
		of the clicked object (`Line2D` or `Rectangle`) in the plot.

		Parameter:
			a `matplotlib.backend_bases.Event`
		"""
		ob = event.artist
		ix = -1
		for i, f in enumerate(self.figs):
			if f == ob.figure: ix = i
		if isinstance(ob, Rectangle):
			self.textBox.setText("%s in year %d is: %d"%(
				figTitles[ix], int(ob.get_x()), int(ob.get_height())))
		elif isinstance(ob, Line2D):
			xdata = ob.get_xdata()
			ydata = ob.get_ydata()
			ind = event.ind
			if ix == 4:
				formatString = "%s in date %s is %.2f"
			else:
				formatString = "%s in date %s is %d"
			self.textBox.setText(formatString%(
				figTitles[ix], np.take(xdata, ind)[0].strftime("%d/%m/%Y"),
				np.take(ydata, ind)[0]))

	def updatePlots(self):
		"""Reset the dialog window removing all the previous items
		and create new canvas for the figures.
		"""
		i = 0
		while True:
			item = self.layout().takeAt(i)
			if item is None:
				break
			del item
		if hasattr(self, "canvas"):
			del self.canvas
		self.canvas = []
		for fig in self.figs:
			if fig is not None:
				self.canvas.append(FigureCanvas(fig))
				self.layout().addWidget(self.canvas[-1], int(i/2), i%2)
				self.canvas[-1].mpl_connect("pick_event", self.pickEvent)
				self.canvas[-1].draw()
				i += 1


class PaperStatsPlots(QDialog):
	"""Class that constructs a window
	to show the results of `paperStats`
	"""

	def __init__(self, fig, title=None, parent=None):
		"""Constructor.
		Defines some properties and buttons

		Parameters:
			fig: the figure to be shown
			title (optional): the window title
			parent (default None): the parent object,
				which should have a property lastPaperStats
		"""
		super(PaperStatsPlots, self).__init__(parent)
		self.fig = fig
		if title is not None:
			self.setWindowTitle(title)
		layout = QGridLayout(self)
		layout.setSpacing(1)
		self.setLayout(layout)
		self.updatePlots()

		nlines = 1
		self.layout().addWidget(PBLabel(
			"Click on the line to have more information:"), nlines + 1, 0)

		self.textBox = QLineEdit("")
		self.textBox.setReadOnly(True)
		self.layout().addWidget(self.textBox, nlines + 2, 0, 1, 2)
		self.saveButton = QPushButton('Save', self)
		self.saveButton.clicked.connect(self.saveAction)
		self.layout().addWidget(self.saveButton, nlines + 3, 0)

		self.clButton = QPushButton('Close', self)
		self.clButton.clicked.connect(self.onClose)
		self.clButton.setAutoDefault(True)
		self.layout().addWidget(self.clButton, nlines + 3, 1)

	def onClose(self):
		"""Close dialog"""
		QDialog.close(self)

	def saveAction(self):
		"""Save the plot into a file,
		after asking the directory where to save them
		"""
		savePath = askDirName(self,
			"Where do you want to save the plot of the stats?")
		if savePath != "":
			try:
				self.parent().lastPaperStats["fig"] = pBStats.plotStats(
					paper=True, save=True, path=savePath)
			except AttributeError:
				pBLogger.warning("", exc_info=True)
			infoMessage("Plot saved.")
			self.saveButton.setDisabled(True)

	def pickEvent(self,event):
		"""Save into `self.textBox` the coordinates
		of the clicked object (`Line2D`) in the plot.

		Parameter:
			a `matplotlib.backend_bases.Event`
		"""
		ob = event.artist
		ix = -1
		if isinstance(ob, Line2D):
			xdata = ob.get_xdata()
			ydata = ob.get_ydata()
			ind = event.ind
			formatString = "%s in date %s is %d"
			self.textBox.setText(formatString%("Citations", np.take(
				xdata, ind)[0].strftime("%d/%m/%Y"), np.take(ydata, ind)[0]))

	def updatePlots(self):
		"""Reset the dialog window removing all the previous items
		and create a new canvas for the figure.
		"""
		i = 0
		while True:
			item = self.layout().takeAt(i)
			if item is None:
				break
			del item
		if hasattr(self, "canvas"): del self.canvas
		if self.fig is not None:
			self.canvas = FigureCanvas(self.fig)
			self.layout().addWidget(self.canvas, 0, 0, 1, 2)
			self.canvas.mpl_connect("pick_event", self.pickEvent)
			self.canvas.draw()



"""Module with many classes that are further extended
in the other physbiblio.gui modules.

This file is part of the physbiblio package.
"""
import traceback
import time
from PySide2.QtCore import Qt, Signal, \
	QAbstractItemModel, QAbstractTableModel, QModelIndex, \
	QSortFilterProxyModel, QThread, QUrl
from PySide2.QtGui import QDesktopServices, QPainter, QPixmap
from PySide2.QtWidgets import \
	QAbstractItemView, QAction, QComboBox, QDesktopWidget, QDialog, \
	QGridLayout, QLabel, QLineEdit, QMenu, QStyle, QTableView, \
	QTableWidget, QTableWidgetItem, QVBoxLayout

try:
	from physbiblio.errors import PBErrorManagerClass, pBLogger
	from physbiblio.view import ViewEntry
	from physbiblio.pdf import pBPDF
	from physbiblio.database import pBDB, catString
	import physbiblio.gui.resourcesPyside2
except ImportError:
	print("Could not find physbiblio and its modules!")
	print(traceback.format_exc())


class PBLabel(QLabel):
	"""Extension of `QLabel` with text interaction flags
	which enable text selection with the mouse
	"""

	def __init__(self, *args, **kwargs):
		"""Extend `QLabel.__init__` and call `setTextInteractionFlags`
		to allow text selection with the mouse
		"""
		QLabel.__init__(self, *args, **kwargs)
		self.setTextFormat(Qt.RichText)
		self.setOpenExternalLinks(True)
		self.setTextInteractionFlags(
			Qt.LinksAccessibleByMouse | Qt.TextSelectableByMouse)


class PBLabelRight(PBLabel):
	"""Class that creates a right-aligned QLabel"""

	def __init__(self, label):
		"""The constructor.

		Parameter:
			label: the text label to be passed to QLabel
		"""
		super(PBLabelRight, self).__init__(label)
		self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)


class PBLabelCenter(PBLabel):
	"""Class that creates a center-aligned QLabel"""

	def __init__(self, label):
		"""The constructor.

		Parameter:
			label: the text label to be passed to QLabel
		"""
		super(PBLabelCenter, self).__init__(label)
		self.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)


class PBComboBox(QComboBox):
	"""Personalize QComboBox for faster construction"""

	def __init__(self, parent, fields, current=None):
		"""Constructor.

		Parameters:
			parent: the parent widget
			fields: the list of (string or printable) contents
				to be added to the QComboBox
			current (default None): the value to be set as the initial value
		"""
		super(PBComboBox, self).__init__(parent)
		for f in fields:
			self.addItem("%s"%f)
		if current is not None:
			try:
				self.setCurrentIndex(fields.index(current))
			except ValueError:
				pass


class PBAndOrCombo(PBComboBox):
	"""Shortcut for generating a QComboBox with and/or"""

	def __init__(self, parent, current=None):
		"""Constructor.

		Parameters:
			parent: the parent widget
			current (default None): the value to be set
				as selected at the beginning
		"""
		super(PBAndOrCombo, self).__init__(parent,
			["AND", "OR"], current=current)


class PBTrueFalseCombo(PBComboBox):
	"""Shortcut for generating a QComboBox with true/false"""

	def __init__(self, parent, current=None):
		"""Constructor.

		Parameters:
			parent: the parent widget
			current (default None): the value to be set as selected
				at the beginning
		"""
		super(PBTrueFalseCombo, self).__init__(parent,
			["True", "False"], current=current)


class ObjListWindow(QDialog):
	"""Create a window managing a list (of bibtexs or of experiments)"""

	def __init__(self, parent=None, gridLayout=False):
		"""Init using parent class and create common definitions

		Parameters:
			parent: the parent object
			gridLayout (boolean, default False):
				if True, use a QGridLayout, otherwise a QVBoxLayout
		"""
		super(ObjListWindow, self).__init__(parent)
		self.tableWidth = None
		self.proxyModel = None
		self.gridLayout = gridLayout
		if gridLayout:
			self.currLayout = QGridLayout()
		else:
			self.currLayout = QVBoxLayout()
		self.setLayout(self.currLayout)

	def triggeredContextMenuEvent(self, row, col, event):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def handleItemEntered(self, index):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def cellClick(self, index):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def cellDoubleClick(self, index):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def createTable(self, *args, **kwargs):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def changeFilter(self, string):
		"""Change the filter of the current view.

		Parameter:
			string: the filter string to be matched
		"""
		self.proxyModel.setFilterRegExp(str(string))

	def addFilterInput(self, placeholderText, gridPos=(1, 0)):
		"""Add a `QLineEdit` to change the filter of the list.

		Parameter:
			placeholderText: the text to be shown
				when no filter is present
			gridPos (tuple): if gridLayout is active,
				the position of the `QLineEdit` in the `QGridLayout`
		"""
		self.filterInput = QLineEdit("",  self)
		self.filterInput.setPlaceholderText(placeholderText)
		self.filterInput.textChanged.connect(self.changeFilter)

		if self.gridLayout:
			self.currLayout.addWidget(self.filterInput, *gridPos)
		else:
			self.currLayout.addWidget(self.filterInput)
		self.filterInput.setFocus()

	def setProxyStuff(self, sortColumn, sortOrder):
		"""Prepare the proxy model to filter and sort the view.

		Parameter:
			sortColumn: the index of the column to use
				for sorting at the beginning
			sortOrder: the order for sorting
				(`Qt.AscendingOrder` or `Qt.DescendingOrder`)
		"""
		self.proxyModel = QSortFilterProxyModel(self)
		self.proxyModel.setSourceModel(self.tableModel)
		self.proxyModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
		self.proxyModel.setSortCaseSensitivity(Qt.CaseInsensitive)
		self.proxyModel.setFilterKeyColumn(-1)

		self.tableview = PBTableView(self)
		self.tableview.setModel(self.proxyModel)
		self.tableview.setSortingEnabled(True)
		self.tableview.setMouseTracking(True)
		self.proxyModel.sort(sortColumn, sortOrder)
		self.currLayout.addWidget(self.tableview)

	def finalizeTable(self, gridPos=(1, 0)):
		"""Resize the table to fit the contents,
		connect functions, add to layout

		Parameter:
			gridPos (tuple): if gridLayout is active,
			the position of the `QLineEdit` in the `QGridLayout`
		"""
		self.tableview.resizeColumnsToContents()

		maxh = QDesktopWidget().availableGeometry().height()
		maxw = QDesktopWidget().availableGeometry().width()
		self.setMaximumHeight(maxh)
		self.setMaximumWidth(maxw)

		hwidth = self.tableview.horizontalHeader().length()
		swidth = self.tableview.style().pixelMetric(QStyle.PM_ScrollBarExtent)
		fwidth = self.tableview.frameWidth() * 2

		if self.tableWidth is None:
			if hwidth > maxw - (swidth + fwidth):
				self.tableWidth = maxw - (swidth + fwidth)
			else:
				self.tableWidth = hwidth + swidth + fwidth
		self.tableview.setFixedWidth(self.tableWidth)

		self.setMinimumHeight(600)

		self.tableview.resizeColumnsToContents()
		self.tableview.resizeRowsToContents()

		self.tableview.entered.connect(self.handleItemEntered)
		self.tableview.clicked.connect(self.cellClick)
		self.tableview.doubleClicked.connect(self.cellDoubleClick)

		if self.gridLayout:
			self.currLayout.addWidget(self.tableview, *gridPos)
		else:
			self.currLayout.addWidget(self.tableview)

	def cleanLayout(self):
		"""Delete the previous table widget and other layout items"""
		while True:
			o = self.layout().takeAt(0)
			if o is None: break
			o.widget().deleteLater()

	def recreateTable(self):
		"""Delete the previous table widget and other layout items,
		then create new ones
		"""
		self.cleanLayout()
		self.createTable()


class EditObjectWindow(QDialog):
	"""Create a window for editing or creating an experiment"""

	def __init__(self, parent=None):
		"""Constructor.

		Parameter:
			parent: the parent object
		"""
		super(EditObjectWindow, self).__init__(parent)
		self.textValues = {}
		self.initUI()

	def keyPressEvent(self, e):
		"""Intercept press keys and exit if escape is pressed

		Parameters:
			e: the `PySide2.QtGui.QKeyEvent`
		"""
		if e.key() == Qt.Key_Escape:
			self.onCancel()

	def onCancel(self):
		"""Set that the result should not be considered and exit"""
		self.result	= False
		self.close()

	def onOk(self):
		"""Set that the result should be considered and exit"""
		self.result	= True
		self.close()

	def initUI(self):
		"""Instantiate the `QGridLayout`"""
		self.currGrid = QGridLayout()
		self.currGrid.setSpacing(1)
		self.setLayout(self.currGrid)

	def centerWindow(self):
		"""Use the `QDesktopWidget` to get the relevant information
		and center the widget in the screen.
		"""
		qr = self.frameGeometry()
		cp = QDesktopWidget().availableGeometry().center()
		qr.moveCenter(cp)
		self.move(qr.topLeft())


class PBThread(QThread):
	"""Extend `QThread`, but further extension is needed"""

	finished = Signal()

	def __init__(self,  parent=None):
		"""Construct the class using `QThread.__init__`

		Parameters:
			parent: the parent object (default None)
		"""
		QThread.__init__(self, parent)

	def run(self):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def setStopFlag(self):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def start(self, *args, **kwargs):
		"""Wait 0.3s and then call `QThread.start`
		(pass all the arguments)
		"""
		time.sleep(0.3)
		QThread.start(self, *args, **kwargs)


class WriteStream(PBThread):
	"""Class used to redirect prints to a window"""

	newText = Signal(str)

	finished = Signal()

	def __init__(self, queue, parent=None, *args, **kwargs):
		"""Constructor

		Parameters:
			queue: a Queue instance
			parent (optional): the parent widget
		"""
		super(WriteStream, self).__init__(parent, *args, **kwargs)
		self.queue = queue
		self.running = True

	def write(self, text):
		"""Write the given text

		Parameters:
			text: the text to send to the output stream
		"""
		self.queue.put(text)

	def run(self):
		"""Run the thread"""
		while self.running:
			text = self.queue.get()
			self.newText.emit(text)
		self.finished.emit()


class PBTableView(QTableView):
	"""Extension of `QTableView`, used to define the contextMenuEvent"""

	def contextMenuEvent(self, event):
		"""Connect the context menu event to the parent function

		Parameter:
			event: the `PySide2.QtGui.QContextMenuEvent`
		"""
		self.parent().triggeredContextMenuEvent(
			self.rowAt(event.y()),
			self.columnAt(event.x()),
			event)


class PBTableModel(QAbstractTableModel):
	"""Extension of `QAbstractTableModel`,
	used for experiments and bibtex entries
	"""

	def __init__(self, parent, header, ask=False,  previous=[], *args):
		"""Constructor, based on `QAbstractTableModel.__init__`

		Parameters:
			parent: the parent widget
			header: the list of column names which constitute the table
			ask (boolean, default False):
				when True, allow to select the lines with a checkbox
			previous: the list of lines which must be selected at the beginning
		"""
		QAbstractTableModel.__init__(self, parent, *args)
		self.header = header
		self.parentObj = parent
		self.previous = previous
		self.ask = ask

	def parent(self):
		"""Return the parent object"""
		return self.parentObj

	def changeAsk(self, new=None):
		"""Change the value of `self.ask` to show/hide checkboxes

		Parameters:
			new: `None` to just swap the current value of `self.ask`,
				`True` or `False` to set it to the desired value
		"""
		self.layoutAboutToBeChanged.emit()
		if new is None:
			self.ask = not self.ask
		elif new == True or new == False:
			self.ask = new
		self.layoutChanged.emit()

	def getIdentifier(self, element):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def prepareSelected(self):
		"""Fill the dictionary `self.selectedElements`
		according to previous selection
		"""
		self.layoutAboutToBeChanged.emit()
		self.selectedElements = {}
		try:
			for bib in self.dataList:
				self.selectedElements[self.getIdentifier(bib)] = False
		except AttributeError:
			pBLogger.exception("dataList is not defined!")
		for prevK in self.previous:
			try:
				if self.selectedElements[prevK] == False:
					self.selectedElements[prevK] = True
			except (KeyError,IndexError):
				pBLogger.exception(
					"Invalid identifier in previous selection: %s"%(prevK))
		self.layoutChanged.emit()

	def selectAll(self):
		"""Select all the available rows"""
		self.layoutAboutToBeChanged.emit()
		for key in self.selectedElements.keys():
			self.selectedElements[key] = True
		self.layoutChanged.emit()

	def unselectAll(self):
		"""Unselect all the available rows"""
		self.layoutAboutToBeChanged.emit()
		for key in self.selectedElements.keys():
			self.selectedElements[key] = False
		self.layoutChanged.emit()

	def addImage(self, imagePath, height):
		"""Create a cell containing an image

		Parameters:
			imagePath: the path of the image
			height: the height to give to the image

		Output:
			a QPixmap
		"""
		return QPixmap(imagePath).scaledToHeight(height)

	def addImages(self, imagePaths, outHeight, height = 48):
		"""Create a cell containing multiple images, using a `QPainter`

		Parameters:
			imagePaths: a list of image paths
				for creating the single `QPixmap`s
			outHeight: the final height for rescaling the image
			height (default 48): the height and width
				to be used when painting

		Output:
			a `QPixmap`
		"""
		width = len(imagePaths) * height
		pm = QPixmap(width, height)
		pm.fill(Qt.transparent)
		self.painter = QPainter(pm)
		for i, img in enumerate(imagePaths):
			self.painter.drawPixmap(i * height, 0, QPixmap(img))
		self.painter.end()
		return pm.scaledToHeight(outHeight)

	def rowCount(self, parent = None):
		"""Count the rows of the given model based on the header

		Parameter:
			parent: `QModelIndex` (required by the parent signature)

		Output:
			the number od rows or zero (if error occurred)
		"""
		try:
			return len(self.dataList)
		except (TypeError, AttributeError):
			return 0

	def columnCount(self, parent = None):
		"""Count the columns of the given model based on the header

		Parameter:
			parent: `QModelIndex` (required by the parent signature)

		Output:
			the number of columns or zero (if error occurred)
		"""
		try:
			return len(self.header)
		except TypeError:
			return 0

	def data(self, index, role):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def flags(self, index):
		"""Determine the flags of a given item

		Parameter:
			index: a `QModelIndex`

		Output:
			None if the index is not valid
			if `self.ask` and first column, show checkboxes:
			Qt.ItemIsUserCheckable | Qt.ItemIsEditable |
				Qt.ItemIsEnabled | Qt.ItemIsSelectable
			all the other cases: Qt.ItemIsEnabled | Qt.ItemIsSelectable
		"""
		if not index.isValid():
			return None
		if index.column() == 0 and self.ask:
			return Qt.ItemIsUserCheckable | Qt.ItemIsEditable | \
				Qt.ItemIsEnabled | Qt.ItemIsSelectable
		else:
			return Qt.ItemIsEnabled | Qt.ItemIsSelectable

	def headerData(self, col, orientation, role):
		"""Obtain column name if correctly asked

		Parameters:
			col: the column index in `self.header`
			orientation: int from `Qt.Orientation`
			role: int from `Qt.ItemDataRole`

		Output:
			the column name or `None`
		"""
		if orientation == Qt.Horizontal and role == Qt.DisplayRole:
			return self.header[col]
		return None

	def setData(self, index, value, role):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def sort(self, col=1, order=Qt.AscendingOrder):
		"""Sort table by the given column, number `col`

		Parameters:
			col: the column name
			order: int from `Qt.SortOrder`
		"""
		self.layoutAboutToBeChanged.emit()
		try:
			self.dataList = sorted(self.dataList, key = lambda x: x[col] )
			if order == Qt.DescendingOrder:
				self.dataList.reverse()
		except KeyError:
			pBLogger.warning("Wrong column name in `sort`: '%s'!"%col)
		self.layoutChanged.emit()


#https://www.hardcoded.net/articles/using_qtreeview_with_qabstractitemmodel
class TreeNode(object):
	"""Create an object that will work as a tree node"""

	def __init__(self, parent, row):
		"""Constructor, set basic properties

		Parameters:
			parent: the parent node
			row: the content of the data row
		"""
		self.parentObj = parent
		self.row = row
		self.subnodes = self._getChildren()

	def parent(self):
		"""Return the parent object"""
		return self.parentObj

	def _getChildren(self):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()


class TreeModel(QAbstractItemModel):
	"""Model for a tree structure."""

	def __init__(self):
		"""Class constructor. Calls `_getRootNodes`
		to build the tree structure
		"""
		QAbstractItemModel.__init__(self)
		self.rootNodes = self._getRootNodes()

	def _getRootNodes(self):
		"""Not implemented: requires a subclass"""
		raise NotImplementedError()

	def index(self, row, column, parent=QModelIndex()):
		"""Retrieve the `QModelIndex` of the requested object

		Parameters:
			row (int): the row index
			column (int): the column index
			parent: the parent `QModelIndex`

		Output:
			A `QModelIndex` instance
		"""
		if not isinstance(parent, QModelIndex):
			pBLogger.debug("Invalid parent '%s' in TreeModel.index"%parent,
				exc_info=True)
			return QModelIndex()
		if not parent.isValid():
			try:
				return self.createIndex(row, column, self.rootNodes[row])
			except IndexError:
				return QModelIndex()
		parentNode = parent.internalPointer()
		try:
			return self.createIndex(row, column, parentNode.subnodes[row])
		except IndexError:
			return QModelIndex()

	def parent(self, index):
		"""Retrieve the `QModelIndex` of the parent of
		the item with the given index, if it exists,
		or an invalid `QModelIndex` instead

		Parameters:
			index: the `QModelIndex` of the node

		Output:
			A `QModelIndex` instance
		"""
		if not isinstance(index, QModelIndex):
			pBLogger.debug("Invalid index '%s' in TreeModel.parent"%index,
				exc_info=True)
			return QModelIndex()
		if not index.isValid():
			return QModelIndex()
		nodeParent = index.internalPointer().parent()
		if nodeParent is None:
			return QModelIndex()
		else:
			return self.createIndex(nodeParent.row, 0, nodeParent)

	def rowCount(self, parent=QModelIndex()):
		"""Count the rows in a given tree branch

		Parameter:
			parent: the `QModelIndex` of the branch parent

		Output:
			the line number
		"""
		if not isinstance(parent, QModelIndex):
			pBLogger.debug("Invalid parent '%s' in TreeModel.rowCount"%parent,
				exc_info=True)
			return len(self.rootNodes)
		if not parent.isValid():
			return len(self.rootNodes)
		node = parent.internalPointer()
		return len(node.subnodes)


class NamedElement(object):
	"""Basic object for the tree structure of categories"""

	def __init__(self, idCat, name, subelements):
		"""Class constructor, set some element properties

		Parameters:
			idCat: the category id
			name: the name of the category
			subelements: the list of children elements
		"""
		self.idCat = idCat
		self.name = name
		self.text = catString(idCat, pBDB)
		self.subelements = subelements


class NamedNode(TreeNode):
	"""Extend `TreeNode` to work with `NamedElement`"""

	def __init__(self, element, parent, row):
		"""Define `self.element` and call `TreeNode.__init__`

		Parameters:
			element: the `NamedElement` of the object
			parent: the parent node
			row: the row index
		"""
		self.element = element
		TreeNode.__init__(self, parent, row)

	def _getChildren(self):
		"""Return a list of `NamedNode`s, containing the children nodes.
		Overrides `TreeNode._getChildren`
		"""
		return [NamedNode(elem, self, index)
			for index, elem in enumerate(self.element.subelements)]


#http://gaganpreet.in/blog/2013/07/04/qtreeview-and-custom-filter-models/
class LeafFilterProxyModel(QSortFilterProxyModel):
	"""Class to override the following behaviour:
		If a parent item doesn't match the filter,
		none of its children will be shown.

	This Model matches items which are descendants
	or ascendants of matching items.
	"""

	def filterAcceptsRow(self, row_num, source_parent):
		"""Overriding the parent function

		Parameters:
			row_num: the row number
			source_parent: the parent node in the tree
		"""
		# Check if the current row matches
		if self.filterAcceptsRowItself(row_num, source_parent):
			return True

		# Traverse up all the way to root and check if any of them match
		if self.filterAcceptsAnyParent(source_parent):
			return True

		# Finally, check if any of the children match
		return self.hasAcceptedChildren(row_num, source_parent)

	def filterAcceptsRowItself(self, row_num, parent):
		"""New name for the original `filterAcceptsRow`,
		which has been overridden

		Parameters:
			row_num: the row number
			parent: the parent node in the tree
		"""
		return super(LeafFilterProxyModel, self).filterAcceptsRow(
			row_num, parent)

	def filterAcceptsAnyParent(self, parent):
		"""Traverse to the root node and check if any of the
		ancestors match the filter

		Parameter:
			parent: the parent node in the tree
		"""
		while parent.isValid():
			if self.filterAcceptsRowItself(parent.row(), parent.parent()):
				return True
			parent = parent.parent()
		return False

	def hasAcceptedChildren(self, row_num, parent):
		"""Starting from the current node as root, traverse all
		the descendants and test if any of the children match

		Parameters:
			row_num: the row number
			parent: the parent node in the tree
		"""
		model = self.sourceModel()
		source_index = model.index(row_num, 0, parent)

		children_count =  model.rowCount(source_index)
		for i in range(children_count):
			if self.filterAcceptsRow(i, source_index):
				return True
		return False


class PBDDTableWidget(QTableWidget):
	"""Drag and drop extension of QTableWidget"""

	def __init__(self, parent, header):
		"""Set some properties and settings.

		Parameters:
			header: the title of the column
		"""
		super(PBDDTableWidget, self).__init__(parent)
		self.setColumnCount(1)
		self.setHorizontalHeaderLabels([header])
		self.setDragEnabled(True)
		self.setAcceptDrops(True)
		self.setSelectionBehavior(QAbstractItemView.SelectRows)
		self.setDragDropOverwriteMode(False)

	# Override this method to get the correct row index for insertion
	def dropMimeData(self, row, col, mimeData, action):
		"""Overridden method to get the row index for insertion

		Parameters:
			row: the index of the dropped row
			col, mimeData, action: not used params
				(see the signature of `QTableWidget.dropMimeData`)
		"""
		self.last_drop_row = row
		return True

	def dropEvent(self, event):
		"""Accept dropEvent and move the selected item
		to the new position

		Parameter:
			event: a `QDropEvent`
		"""
		# The QTableWidget from which selected rows will be moved
		sender = event.source()

		# Default dropEvent method fires dropMimeData
		# with appropriate parameters (we're interested in the row index).
		super(PBDDTableWidget, self).dropEvent(event)
		# Now we know where to insert selected row(s)
		dropRow = self.last_drop_row

		selectedRows = sender.getselectedRowsFast()

		# Allocate space for transfer
		for _ in selectedRows:
			self.insertRow(dropRow)

		# if sender == receiver (self),
		# after creating new empty rows selected rows might change
		# their locations
		sel_rows_offsets = [0 if self != sender or srow < dropRow \
			else len(selectedRows) for srow in selectedRows]
		selectedRows = [row + offset for row, offset \
			in zip(selectedRows, sel_rows_offsets)]

		# copy content of selected rows into empty ones
		for i, srow in enumerate(selectedRows):
			for j in range(self.columnCount()):
				item = sender.item(srow, j)
				if item:
					source = QTableWidgetItem(item)
					self.setItem(dropRow + i, j, source)

		# delete selected rows
		for srow in reversed(selectedRows):
			sender.removeRow(srow)

		event.accept()

	def getselectedRowsFast(self):
		"""Return the list of selected rows

		Output:
			a list with the row indexes of the selected rows
		"""
		selectedRows = []
		for item in self.selectedItems():
			row = item.row()
			text = item.text()
			if row not in selectedRows and text != "bibkey":
				selectedRows.append(row)
		selectedRows.sort()
		return selectedRows


class PBMenu(QMenu):
	"""Extend `QMenu` for faster menu build"""

	def __init__(self, parent=None):
		"""Construct the element defining some basic properties

		Parameter:
			parent: the parent widget
		"""
		super(PBMenu, self).__init__(parent)
		self.possibleActions = []
		self.result = False

	def fillMenu(self):
		"""Add actions, separators or submenus according to the content
		of `self.possibleActions` (recursively for submenus)
		"""
		for act in self.possibleActions:
			if act is None:
				self.addSeparator()
			elif isinstance(act, list):
				submenu = PBMenu()
				submenu.setTitle(act[0])
				submenu.possibleActions = act[1]
				submenu.fillMenu()
				self.addMenu(submenu)
			elif isinstance(act, QAction):
				self.addAction(act)

	def keyPressEvent(self, e):
		"""Intercept press keys and exit if escape is pressed

		Parameters:
			e: the `PySide2.QtGui.QKeyEvent`
		"""
		if e.key() == Qt.Key_Escape:
			self.close()


class GUIViewEntry(ViewEntry):
	"""Extends the ViewEntry class to work with QtGui.QDesktopServices"""

	def openLink(self, key, arg="", fileArg=None):
		"""Use `QDesktopServices` to open an url using
		the system default applications

		Parameters:
			key: the entry key or the link (if `arg` == "link")
			arg:
				if `arg` == "file", `fileArg` must be the file name
				if `arg` == "link", `key` must be the link to be opened
				for any other values, the link will be generated using
					the `physbiblio.view.viewWntry.getLink` method
			fileArg: the file name if `arg` == "file", or
				the argument passed to `physbiblio.view.viewWntry.getLink`
				if needed
		"""
		if isinstance(key, list):
			for k in key:
				self.openLink(k, arg, fileArg)
		else:
			if arg is "file":
				url = QUrl.fromLocalFile(fileArg)
			elif arg is "link":
				url = QUrl(key)
			else:
				link = self.getLink(key, arg = arg, fileArg = fileArg)
				url = QUrl(link)
			if QDesktopServices.openUrl(url):
				pBLogger.debug("Opening link '%s' for entry '%s' successful!"%(
					url.toString(), key))
			else:
				pBLogger.warning("Opening link for '%s' failed!"%key)


pBGuiView = GUIViewEntry()


class PBImportedTableModel(PBTableModel):
	"""Extend `PBTableModel` to manage the selection
	during the import of entries
	"""

	def __init__(self, parent, bibdict, header, idName="ID", *args):
		"""Set some properties and settings

		Parameters:
			parent: the parent widget (pass to `PBTableModel.__init__`)
			bibdict: a dictionary with the info
				of the imported bibtex entries.
				It should contain dictionaries with two items:
					"bibpars", "exist"
			header: the header names to be passed
				to `PBTableModel.__init__`
			idName: the key of the field representing
				the unique ID of the item
		"""
		self.typeClass = "imports"
		self.idName = idName
		self.bibsOrder = [k for k in bibdict.keys()]
		self.dataList = [bibdict[k]['bibpars'] for k in self.bibsOrder]
		self.existList = [bibdict[k]['exist'] for k in self.bibsOrder]
		PBTableModel.__init__(self, parent, header, *args)
		self.prepareSelected()

	def getIdentifier(self, element):
		"""Return the unique identifier of the given element

		Parameters:
			element: a dictionary
		"""
		return element[self.idName]

	def data(self, index, role):
		"""Return the data for the requested cell and role

		Parameters:
			index: a `QModelIndex`
			role: the requested role for the given cell
		"""
		if not index.isValid():
			return None
		row = index.row()
		column = index.column()
		try:
			value = self.dataList[row][self.header[column]]
		except (IndexError, KeyError):
			pBLogger.warning("Missing element", exc_info=True)
			return None

		if role == Qt.CheckStateRole and column == 0 and \
				self.existList[row] is False:
			if self.selectedElements[self.dataList[row][self.idName]] == False:
				return Qt.Unchecked
			else:
				return Qt.Checked
		if role in [Qt.EditRole, Qt.DisplayRole] and column == 0 \
				and self.existList[row] is True:
			return value + " - already existing"
		if role == Qt.EditRole:
			return value
		if role == Qt.DisplayRole:
			return value
		return None

	def setData(self, index, value, role):
		"""Set the selection data for a given index

		Parameters:
			index: a `QModelIndex`
			value: `Qt.Checked` to set the row as selected
			role: `Qt.CheckStateRole` for performing an action,
				or any other role
		"""
		if not index.isValid():
			return False
		if role == Qt.CheckStateRole and index.column() == 0:
			if value == Qt.Checked:
				self.selectedElements[self.dataList[index.row()][self.idName]] \
					= True
			else:
				self.selectedElements[self.dataList[index.row()][self.idName]] \
					= False
		self.dataChanged.emit(index, index)
		return True

	def flags(self, index):
		"""Return the flags for the requested row

		Parameters:
			index: a `QModelIndex`
		"""
		if not index.isValid():
			return None
		if index.column() == 0 and self.existList[index.row()] is False:
			return Qt.ItemIsUserCheckable | Qt.ItemIsEditable | \
				Qt.ItemIsEnabled | Qt.ItemIsSelectable
		else:
			return Qt.ItemIsEnabled | Qt.ItemIsSelectable



#!/usr/bin/env python
"""Test file for the physbiblio.export module.

This file is part of the physbiblio package.
"""
import sys
import traceback
import os
from stat import S_IREAD, S_IRGRP, S_IROTH
import bibtexparser

if sys.version_info[0] < 3:
	import unittest2 as unittest
	from mock import patch
else:
	import unittest
	from unittest.mock import patch

try:
	from physbiblio.setuptests import *
	from physbiblio.config import pbConfig
	from physbiblio.database import pBDB
	from physbiblio.export import pBExport
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
	print(traceback.format_exc())

@unittest.skipIf(skipTestsSettings.long, "Long tests")
class TestExportMethods(unittest.TestCase):
	"""Tests for methods in physbiblio.export"""
	def test_backup(self):
		"""Test backup file creation and related"""
		emptyFileName = os.path.join(pbConfig.dataPath,
			"tests_%s.bib"%today_ymd)
		if os.path.exists(emptyFileName):
			os.remove(emptyFileName)
		if os.path.exists(emptyFileName + pBExport.backupExtension):
			os.remove(emptyFileName + pBExport.backupExtension)
		self.assertFalse(pBExport.backupCopy(emptyFileName))
		self.assertFalse(os.path.exists(emptyFileName))
		self.assertFalse(os.path.exists(
			emptyFileName + pBExport.backupExtension))
		open(emptyFileName, 'a').close()
		self.assertTrue(pBExport.backupCopy(emptyFileName))
		self.assertTrue(os.path.exists(
			emptyFileName + pBExport.backupExtension))
		os.remove(emptyFileName)
		self.assertFalse(os.path.exists(emptyFileName))
		self.assertTrue(pBExport.restoreBackupCopy(emptyFileName))
		self.assertTrue(os.path.exists(emptyFileName))
		os.chmod(emptyFileName, S_IREAD)
		self.assertFalse(pBExport.restoreBackupCopy(emptyFileName))
		os.remove(emptyFileName)
		self.assertTrue(pBExport.rmBackupCopy(emptyFileName))
		self.assertFalse(os.path.exists(
			emptyFileName + pBExport.backupExtension))

	def test_offlineExports(self):
		"""Test of offline export functions exportSelected,
		updateExportedBib
		"""
		testBibName = os.path.join(pbConfig.dataPath,
			"tests_%s.bib"%today_ymd)
		sampleList = [
			{"bibtex": '@Article{empty,\nauthor="me",\ntitle="no"\n}'},
			{"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'}]
		sampleTxt = '@Article{empty,\nauthor="me",\ntitle="no"\n}\n' \
			+ '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}\n'
		pBDB.bibs.lastFetched = sampleList
		pBExport.exportLast(testBibName)
		with open(testBibName) as f:
			self.assertEqual(f.read(), sampleTxt)
		os.remove(testBibName)

		sampleList = [
			{"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="{yes}",\n}'}]
		sampleTxt = '@Article{empty2,\nauthor="me2",\ntitle="{yes}",\n}\n'
		pBExport.exportSelected(testBibName, sampleList)
		with open(testBibName) as f:
			self.assertEqual(f.read(), sampleTxt)

		with patch('physbiblio.database.Entries.getByBibkey',
				return_value=[{"bibtexDict":
					{"ID": "empty2", "ENTRYTYPE": "Article",
					"author": "me et al", "title": "yes"}}]) as _mock:
			pBExport.updateExportedBib(testBibName, overwrite=True)
		with open(testBibName) as f:
			self.assertEqual(f.read().replace(" ","").replace("\n",""),
			 sampleTxt.replace("me2", 'me et al').replace(" ",""
				).replace("\n",""))
		pBExport.rmBackupCopy(testBibName)
		os.remove(testBibName)

	def test_exportAll(self):
		"""Test of exportAll"""
		testBibName = os.path.join(pbConfig.dataPath,
			"tests_%s.bib"%today_ymd)
		sampleList = [
			{"bibtex": '@Article{empty,\nauthor="me",\ntitle="no"\n}'},
			{"bibtex": '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'}]
		sampleTxt = '@Article{empty,\nauthor="me",\ntitle="no"\n}\n' \
			+ '@Article{empty2,\nauthor="me2",\ntitle="yes"\n}\n'
		with patch('physbiblio.database.Entries.fetchCursor',
				return_value=sampleList) as _curs:
			pBExport.exportAll(testBibName)
		with open(testBibName) as f:
			self.assertEqual(f.read(), sampleTxt)
		os.remove(testBibName)

	def test_exportForTexFile(self):
		"""test exportForTexFile function with a fake tex and database"""
		testBibName = os.path.join(pbConfig.dataPath, "tests_%s.bib"%today_ymd)
		testTexName = os.path.join(pbConfig.dataPath, "tests_%s.tex"%today_ymd)
		texString = "\cite{empty,prova, empty2}\citep{empty2}" \
			+ "\citet{Gariazzo:2015rra}, \citet{Gariazzo:2017rra}\n"
		with open(testTexName, "w") as f:
			f.write(texString)
		with open(testTexName) as f:
			self.assertEqual(f.read(), texString)
		sampleList = [{"bibkey": "empty", "bibtex":
			'@Article{empty,\n        author = "me",\n         '
			+ 'title = "{no}",\n}'},
			{"bibkey": "empty2", "bibtex":
				'@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'}]

		self.assertFalse(os.path.exists(testBibName))
		with patch('physbiblio.database.CatsEntries.insert',
				return_value=True) as _ceins:
			with patch('physbiblio.database.Entries.getByBibtex',
					side_effect=[
					[{"bibkey": "empty", "bibtexDict": {}, "bibtex":
						'@Article{empty,\nauthor="me",\ntitle="no"\n}'}],
					[],
					[{"bibkey": "Gariazzo:2015rra", "bibtexDict": {},
					"bibtex": '@article{Gariazzo:2015rra,\nauthor= '
					+ '"Gariazzo, S. and others",\ntitle='
					+ '"{Light sterile neutrinos}",\n}'}],
					[{"bibkey": "empty2", "bibtexDict": {}, "bibtex":
						'@Article{empty2,\nauthor="me2",\ntitle="yes"\n}'}],
					[],
					[{"bibkey": "Gariazzo:2015rra", "bibtexDict": {},
					"bibtex": '@article{Gariazzo:2015rra,\nauthor= '
					+ '"Gariazzo, S. and others",\ntitle='
					+ '"{Light sterile neutrinos}",\n}'}],
					[],
					[],
					]) as _getbbibt:
				with patch('physbiblio.database.Entries.loadAndInsert',
						side_effect=[
						'@article{Gariazzo:2015rra,\nauthor= '
						+ '"Gariazzo, S. and others",\ntitle='
						+ '"{Light sterile neutrinos}",\n}',
						'@article{Gariazzo:2015rra,\nauthor= '
						+ '"Gariazzo, S. and others",\ntitle='
						+ '"{Light sterile neutrinos}",\n}',
						'']) as _mock:
					output = pBExport.exportForTexFile(
						testTexName, testBibName, overwrite=True,
						autosave=False)
		self.assertEqual(output[0],
			['empty', 'prova', 'empty2', 'Gariazzo:2015rra',
			'Gariazzo:2017rra']) #requiredBibkeys
		self.assertEqual(output[1],
			['prova', 'Gariazzo:2015rra', 'Gariazzo:2017rra']) #missing
		self.assertEqual(output[2], ['Gariazzo:2015rra']) #retrieved
		self.assertEqual(output[3], ['Gariazzo:2017rra']) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {'prova': 'Gariazzo:2015rra'}) #newKeys
		self.assertEqual(output[6], 2) #warnings
		self.assertEqual(output[7], 5) #total
		self.assertTrue(os.path.exists(testBibName))
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{empty,\n        author = "me",\n         title = '
					+ '"{no}",\n}\n\n',
				'@Article{prova,\n        author = "Gariazzo, S. and others",'
					+ '\n         title = "{Light sterile neutrinos}",'
					+ '\n}\n\n',
				'@Article{empty2,\n        author = "me2",\n         title = '
					+ '"{yes}",\n}\n\n',
				'@Article{Gariazzo:2015rra,\n        author = '
					+ '"Gariazzo, S. and others",\n         title = '
					+ '"{Light sterile neutrinos}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		with open(testTexName, "a") as f:
			f.write("\cite{newcite}")
		with patch('physbiblio.database.CatsEntries.insert',
				return_value=True) as _ceins, \
				patch('physbiblio.database.Entries.getByBibtex',
					side_effect=[
					[],
					[],
					[],
					[{"bibkey": "newcite", "bibtexDict": {}, "bibtex":
						'@article{newcite,\nauthor= "myself",\ntitle='
						+ '"{some paper}",\n}'}],
					]) as _getbbibt,\
				patch('physbiblio.database.Entries.loadAndInsert',
					side_effect=['',
					'@article{newcite,\nauthor= "myself",\ntitle='
						+ '"{some paper}",\n}']) as _mock:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, autosave=False)
		self.assertEqual(output[0],
			['Gariazzo:2017rra', 'newcite']) #requiredBibkeys
		self.assertEqual(output[1], ['Gariazzo:2017rra', 'newcite']) #missing
		self.assertEqual(output[2], ['newcite']) #retrieved
		self.assertEqual(output[3], ['Gariazzo:2017rra']) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {}) #newKeys
		self.assertEqual(output[6], 1) #warnings
		self.assertEqual(output[7], 6) #total
		self.assertTrue(os.path.exists(testBibName))
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{empty,\n        author = "me",\n         title = '
					+ '"{no}",\n}\n\n',
				'@Article{prova,\n        author = "Gariazzo, S. and others"'
					+ ',\n         title = "{Light sterile neutrinos}",\n}\n\n',
				'@Article{empty2,\n        author = "me2",\n         title = '
					+ '"{yes}",\n}\n\n',
				'@Article{Gariazzo:2015rra,\n        author = '
					+ '"Gariazzo, S. and others",\n         title = '
					+ '"{Light sterile neutrinos}",\n}\n\n',
				'@Article{newcite,\n        author = "myself",'
					+ '\n         title = "{some paper}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		with open(testTexName, "w") as f:
			f.write("\cite{newcite:now18}")
		with patch('physbiblio.database.Entries.getByBibtex', side_effect=[
				[{"bibkey": "newcite:now18", "bibtexDict": {},
				"bibtex": '@article{newcite:now18,\nauthor= "myself",\ntitle='
					+ '"{some paper}",\n}'}],
				]) as _getbbibt:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, overwrite=True, autosave=False)
		self.assertEqual(output[0], ['newcite:now18']) #requiredBibkeys
		self.assertEqual(output[1], []) #missing
		self.assertEqual(output[2], []) #retrieved
		self.assertEqual(output[3], []) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {}) #newKeys
		self.assertEqual(output[6], 0) #warnings
		self.assertEqual(output[7], 1) #total
		self.assertTrue(os.path.exists(testBibName))
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{newcite:now18,\n        author = "myself",\n         '
					+ 'title = "{some paper}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		with open(testTexName, "w") as f:
			f.write("\cite{newcite:now18}\cite{bib1}\cite{bib2}")
		bibtex1 = '@Article{newcite,\nauthor="S. Gariazzo",\ntitle=' \
			+ '"{newtitle}"\n}'
		bibtex2 = '@article{bib1,\nauthor="SG",\ntitle=' \
			+ '"{Light sterile neutrinos}",\n}'
		bibtex3 = '@Article{bib2,\nauthor="me",\ntitle="title"\n}'
		with patch('physbiblio.database.Entries.getByBibtex', side_effect=[
				[{"bibkey": "newcite:now18", "bibtexDict":
					bibtexparser.loads(bibtex1).entries[0],
					"bibtex": bibtex1}],
				[{"bibkey": "bib1", "bibtexDict":
					bibtexparser.loads(bibtex2).entries[0],
					"bibtex": bibtex2}],
				[{"bibkey": "bib2", "bibtexDict":
					bibtexparser.loads(bibtex3).entries[0],
					"bibtex": bibtex3}],
				]) as _getbbibt:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, autosave=False, updateExisting=True)
		self.assertEqual(output[0], ['bib1', 'bib2']) #requiredBibkeys
		self.assertEqual(output[1], []) #missing
		self.assertEqual(output[2], []) #retrieved
		self.assertEqual(output[3], []) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {}) #newKeys
		self.assertEqual(output[6], 0) #warnings
		self.assertEqual(output[7], 3) #total
		self.assertTrue(os.path.exists(testBibName))
		self.maxDiff = None
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{newcite:now18,\n        author = "S. Gariazzo",'
					+ '\n         title = "{newtitle}",\n}\n\n',
				'@Article{bib1,\n        author = "SG",\n         title = '
					+ '"{Light sterile neutrinos}",\n}\n\n',
				'@Article{bib2,\n        author = "me",\n         title = '
					+ '"{title}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		with open(testTexName, "w") as f:
			f.write("\cite{newcite:now18}\cite{bib1}")
		bibtex1 = '@Article{newcite:now18,\nauthor="S. Gariazzo",\ntitle=' \
			+ '"{newtitle}"\n}'
		bibtex2 = '@article{bib1,\nauthor="SG",\ntitle=' \
			+ '"{Light sterile neutrinos}",\n}'
		with patch('physbiblio.database.Entries.getByBibtex', side_effect=[
				[{"bibkey": "newcite:now18", "bibtexDict":
					bibtexparser.loads(bibtex1).entries[0],
					"bibtex": bibtex1}],
				[{"bibkey": "bib1", "bibtexDict":
					bibtexparser.loads(bibtex2).entries[0],
					"bibtex": bibtex2}],
				]) as _getbbibt:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, autosave=False, removeUnused=True)
		self.assertEqual(output[0], []) #requiredBibkeys
		self.assertEqual(output[1], []) #missing
		self.assertEqual(output[2], []) #retrieved
		self.assertEqual(output[3], []) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {}) #newKeys
		self.assertEqual(output[6], 0) #warnings
		self.assertEqual(output[7], 2) #total
		self.assertTrue(os.path.exists(testBibName))
		self.maxDiff = None
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{newcite:now18,\n        author = "S. Gariazzo",'
					+ '\n         title = "{newtitle}",\n}\n\n',
				'@Article{bib1,\n        author = "SG",\n         '
					+ 'title = "{Light sterile neutrinos}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		with open(testTexName, "w") as f:
			f.write("\cite{newcite}")
		bibtex1 = '@Article{newcite,\nauthor="S. Gariazzo",' \
			+ '\ntitle="{title}"\n}'
		with patch('physbiblio.database.Entries.getByBibtex', side_effect=[
				[{"bibkey": "newcite", "bibtexDict":
					bibtexparser.loads(bibtex1).entries[0],
					"bibtex": bibtex1}],
				]) as _getbbibt:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, autosave=False, overwrite=True)
		self.assertEqual(output[0], ['newcite']) #requiredBibkeys
		self.assertEqual(output[1], []) #missing
		self.assertEqual(output[2], []) #retrieved
		self.assertEqual(output[3], []) #notFound
		self.assertEqual(output[4], []) #unexpected
		self.assertEqual(output[5], {}) #newKeys
		self.assertEqual(output[6], 0) #warnings
		self.assertEqual(output[7], 1) #total
		self.assertTrue(os.path.exists(testBibName))
		self.maxDiff = None
		with open(testBibName) as f:
			newTextBib = f.read()
		for t in ['%file written by PhysBiblio\n',
				'@Article{newcite,\n        author = "S. Gariazzo",'
				+ '\n         title = "{title}",\n}\n\n']:
			self.assertIn(t, newTextBib)

		os.remove(testBibName)
		with patch("logging.Logger.error") as _er:
			output = pBExport.exportForTexFile(testTexName,
				testBibName, autosave=False)
			_er.assert_any_call(
				"Cannot read file %s.\nCreating one."%testBibName)
		self.assertEqual(len(output), 8)
		self.assertTrue(os.path.exists(testBibName))
		with patch("logging.Logger.exception") as _ex:
			self.assertFalse(pBExport.exportForTexFile(testTexName,
				"/surely/not/existing/path.bib", autosave=False))
			_ex.assert_called_once_with(
				'Cannot create file /surely/not/existing/path.bib!')
		os.remove(testBibName)
		os.remove(testTexName)


if __name__=='__main__':
	unittest.main()



"""Module for retrieving bibtex information
from a list of given web resources
(see `__all__`).

This module is part of the physbiblio package.
"""
__author__ = 'Stefano Gariazzo'
__email__ = "stefano.gariazzo@gmail.com"

import physbiblio
__version__ = physbiblio.__version__
__version_date__ = physbiblio.__version_date__

__all__ = [
	"arxiv",
	"doi",
	"inspire",
	"inspireoai",
	"isbn",
	"webInterf",
	"tests",
	]



"""The package that contains the tests for the physbiblio package.

This file is part of the physbiblio package.
"""
__author__ = 'Stefano Gariazzo'
__email__ = "stefano.gariazzo@gmail.com"

import physbiblio
__version__ = physbiblio.__version__
__version_date__ = physbiblio.__version_date__

__all__ = [
	"test_argParser",
	"test_bibtexWriter",
	"test_config",
	"test_database",
	"test_errors",
	"test_export",
	"test_inspireStats",
	"test_pdf",
	"test_view",
	]




"""Module that deals with importing info from the arXiv API.

Uses feedparser module to read the page content.

This file is part of the physbiblio package.
"""
import sys
import re
import traceback
import feedparser
try:
	from physbiblio.config import pbConfig
	from physbiblio.errors import pBLogger
	from bibtexparser.bibdatabase import BibDatabase
	from physbiblio.webimport.webInterf import WebInterf, physBiblioWeb
	from physbiblio.parseAccents import parse_accents_str
	from physbiblio.bibtexWriter import pbWriter
except ImportError:
	print("Could not find physbiblio and its modules!")
	print(traceback.format_exc())
	raise


def getYear(string):
	"""Use the arxiv id to compute the year"""
	identif = re.compile("([0-9]{2})([0-9]{2}.[0-9]{4,5}|[0-9]{5})")
	try:
		for t in identif.finditer(string):
			if len(t.group()) > 0:
				a = t.group(1)
				if int(a) > 90:
					return "19" + a
				else:
					return "20" + a
	except Exception:
		pBLogger.warning("Error in converting year from '%s'"%string)
		return None


class WebSearch(WebInterf):
	"""Subclass of WebInterf that can connect to arxiv.org
	to perform searches
	"""

	def __init__(self):
		"""Initializes the class variables using
		the WebInterf constructor.

		Define additional specific parameters for the arxiv.org API.
		"""
		WebInterf.__init__(self)
		self.name = "arXiv"
		self.description = "arXiv fetcher"
		self.url = "https://export.arxiv.org/api/query"
		self.urlRss = "https://export.arxiv.org/rss/"
		self.urlArgs = {
			"start":"0"}
		self.categories = {
			"astro-ph": ["CO", "EP", "GA", "HE", "IM", "SR"],
			"cond-mat": ["dis-nn", "mes-hall", "mtrl-sci", "other",
				"quant-gas", "soft", "stat-mech", "str-el", "supr-con"],
			"cs": ["AI", "AR", "CC", "CE", "CG", "CL", "CR", "CV",
				"CY", "DB", "DC", "DL", "DM", "DS", "ET", "FL", "GL",
				"GR", "GT", "HC", "IR", "IT", "LG", "LO", "MA", "MM",
				"MS", "NA", "NE", "NI", "OH", "OS", "PF", "PL", "RO",
				"SC", "SD", "SE", "SI", "SY"],
			"econ": ["EM", "GN", "TH"],
			"eess": ["AS", "IV", "SP"],
			"gr-qc": [],
			"hep-ex": [],
			"hep-lat": [],
			"hep-ph": [],
			"hep-th": [],
			"math": ["AC", "AG", "AP", "AT", "CA", "CO", "CT", "CV", "DG",
				"DS", "FA", "GM", "GN", "GR", "GT", "HO", "IT", "KT", "LO",
				"MG", "MP", "NA", "NT", "OA", "OC", "PR", "QA", "RA", "RT",
				"SG", "SP", "ST"],
			"math-ph": [],
			"nlin": ["AO", "CD", "CG", "PS", "SI"],
			"nucl-ex": [],
			"nucl-th": [],
			"physics": ["acc-ph", "ao-ph", "app-ph", "atm-clus", "atom-ph",
				"bio-ph", "chem-ph", "class-ph", "comp-ph", "data-an",
				"ed-ph", "flu-dyn", "gen-ph", "geo-ph", "hist-ph", "ins-det",
				"med-ph", "optics", "plasm-ph", "pop-ph", "soc-ph", "space-ph"],
			"q-bio": [
				"BM", "CB", "GN", "MN", "NC", "OT", "PE", "QM", "SC", "TO"],
			"q-fin": ["CP", "EC", "GN", "MF", "PM", "PR", "RM", "ST", "TR"],
			"quant-ph": [],
			"stat": ["AP", "CO", "ME", "ML", "OT", "TH"]
			}

	def getYear(self, string):
		"""Use the arxiv id to compute the year"""
		return getYear(string)

	def retrieveUrlFirst(self, string, searchType="all", **kwargs):
		"""Retrieves the first result from the content
		of the given web page.
		The function calls arxivRetriever which will do the job.

		Parameters:
			string: the search string
			searchType: the search method in arxiv API (default 'all')

		Output:
			returns the bibtex string built from arxivRetriever
		"""
		return self.arxivRetriever(string,
			searchType, additionalArgs={"max_results":"1"}, **kwargs)

	def retrieveUrlAll(self, string, searchType="all", **kwargs):
		"""Retrieves all the results from the content
		of the given web page.
		The function calls arxivRetriever which will do the job.

		Parameters:
			string: the search string
			searchType: the search method in arxiv API (default 'all')

		Output:
			returns the bibtex string built from arxivRetriever
		"""
		return self.arxivRetriever(string,
			searchType,
			additionalArgs = {
				"max_results": pbConfig.params["maxArxivResults"]},
			**kwargs)

	def arxivRetriever(self,
			string,
			searchType="all",
			additionalArgs=None,
			fullDict=False):
		"""Reads the feed content got from arxiv into a dictionary,
		used to return a bibtex.

		Parameters:
			string: the search string
			searchType: the search method in arxiv API (default 'all').
			The possible values are:
				ti->	Title
				au	->	Author
				abs	->	Abstract
				co	->	Comment
				jr	->	Journal Reference
				cat	->	Subject Category
				rn	->	Report Number
				id	->	Id (use id_list instead)
				all	->	All of the above
			additionalArgs: a dictionary of additional arguments
				that can be passed to self.urlArgs (default None)
			fullDict (logical): return the bibtex dictionary in addition
				to the bibtex text (default False)

		Output:
			the bibtex text
			(optional, depending on fullDict): the bibtex Dictionary
		"""
		if additionalArgs:
			for k, v in additionalArgs.items():
				self.urlArgs[k] = v
		self.urlArgs["search_query"] = searchType + ":" + string
		url = self.createUrl()
		pBLogger.info("Search '%s:%s' -> %s"%(searchType, string, url))
		text = parse_accents_str(self.textFromUrl(url))
		try:
			data = feedparser.parse(text)
			db = BibDatabase()
			db.entries = []
			dictionaries = []
			for entry in data['entries']:
				dictionary = {}
				idArx = entry['id'].replace(
					"http://arxiv.org/abs/", "").replace(
					"https://arxiv.org/abs/", "")
				pos = idArx.find("v")
				if pos >= 0:
					idArx = idArx[0:pos]
				dictionary["ENTRYTYPE"] = "article"
				dictionary["ID"] = idArx
				dictionary["archiveprefix"] = "arXiv"
				dictionary["title"] = entry['title']
				dictionary["arxiv"] = idArx
				try:
					dictionary["doi"] = entry['arxiv_doi']
				except KeyError as e:
					pBLogger.debug("KeyError: %s"%e)
				dictionary["abstract"] = entry['summary'].replace("\n", " ")
				dictionary["authors"] = " and ".join([au["name"] \
					for au in entry['authors']])
				dictionary["primaryclass"] = \
					entry['arxiv_primary_category']['term']
				year = self.getYear(dictionary["arxiv"])
				if year is not None:
					dictionary["year"] = year
				db.entries.append(dictionary)
				dictionaries.append(dictionary)
			if fullDict:
				dictionary = dictionaries[0]
				for d in dictionaries:
					if string in d["arxiv"]:
						dictionary = d
				return pbWriter.write(db), dictionary
			else:
				return pbWriter.write(db)
		except Exception:#intercept all other possible errors
			pBLogger.exception("Impossible to get results")
			if fullDict:
				return "",  {}
			else:
				return ""

	def arxivDaily(self, category):
		"""Read daily RSS feed for a given category

		Parameter:
			category: the selected category (see `self.categories)
		"""
		if "." in category:
			main, sub = category.split(".")
		else:
			main = category
			sub = ""
		url = self.urlRss
		if main not in self.categories.keys():
			pBLogger.warning("Main category not found: %s"%main)
			return False
		else:
			url += main
		if sub != "" and sub not in self.categories[main]:
			pBLogger.warning("Sub category not found: %s"%sub)
			return False
		elif sub != "" and sub in self.categories[main]:
			url += "." + sub
		pBLogger.info(url)
		text = self.textFromUrl(url)
		if text is None:
			pBLogger.warning("Url is empty!")
			return False
		author = re.compile('(>|&gt;)([^/]*)(</a>|&lt;/a&gt;)')
		additionalInfo = re.compile(
			' \(arXiv:([0-9\.v]*) \[([\-\.a-zA-Z]*)\]([ A-Z]*)\)')
		if sys.version_info[0] < 3:
			text = text.decode("utf-8")
		try:
			data = feedparser.parse(parse_accents_str(text))
			entries = []
			for element in data.entries:
				tmp = {}
				tmp["eprint"] = element["id"].split("/")[-1]
				tmp["abstract"] = element["summary"].replace(
					"\n", " ").replace("<p>", "").replace("</p>", "")
				tmp["authors"] = [ m.group(2) for m in author.finditer(
					element["authors"][0]["name"]) if m != "" ]
				tmp["author"] = " and ".join(tmp["authors"]) if len(
					tmp["authors"]) < pbConfig.params["maxAuthorNames"] \
					else " and ".join(
						tmp["authors"][0:pbConfig.params["maxAuthorNames"]] \
						+ ["others"])
				tmp["replacement"] = "UPDATED" in element["title"]
				tmp["primaryclass"] = [m.group(2) for m in \
					additionalInfo.finditer(element["title"]) if m != ""][0]
				tmp["cross"] = "CROSS LISTED" in element["title"] or \
					category.lower() not in tmp["primaryclass"].lower()
				tmp["version"] = [m.group(1) for m in additionalInfo.finditer(
					element["title"]) if m != ""][0]
				parenthesis = [m.group() for m in additionalInfo.finditer(
					element["title"]) if m != ""][0]
				tmp["title"] = element["title"].replace(parenthesis, "")
				entries.append(tmp)
			return entries
		except Exception:
			pBLogger.error("Cannot parse arxiv RSS feed:\n%s"%text,
				exc_info=True)
			return False



"""Module that manages the actions on the database (and few more).

This file is part of the physbiblio package.
"""
import traceback
import sqlite3
from sqlite3 import \
	OperationalError, ProgrammingError, DatabaseError, InterfaceError
import os
import ast

try:
	import physbiblio.tablesDef
except ImportError:
	print("Could not find physbiblio and its modules!")
	print(traceback.format_exc())
	raise

encoding_default = 'iso-8859-15'


class PhysBiblioDBCore():
	"""Contains most of the basic functions on the database.
	Will be subclassed to do everything else.
	"""

	def __init__(self, dbname, logger, noOpen=False, info=True):
		"""Initialize database class (column names, descriptions)
		and opens the database.

		Parameters:
			dbname: the name of the database to be opened
			logger: the logger used for messages
			noOpen (boolean, default False): open the database or not
			info (boolean, default True): show some output when opening DB
		"""
		#structure of the tables
		self.tableFields = physbiblio.tablesDef.tableFields
		self.descriptions = physbiblio.tablesDef.fieldsDescriptions
		#names of the columns
		self.tableCols = {}
		for q in self.tableFields.keys():
			self.tableCols[q] = [ a[0] for a in self.tableFields[q] ]

		self.dbChanged = False
		self.conn = None
		self.curs = None
		self.dbname = dbname
		self.logger = logger
		db_is_new = not os.path.exists(self.dbname)

		if not noOpen:
			self.openDB(info=info)
			if db_is_new or self.checkExistingTables():
				self.logger.info("-------New database or missing tables."
					+ "Creating them!\n\n")
				self.createTables()
			self.checkCols()

		self.lastFetched = None
		self.catsHier = None

		self.loadSubClasses()

	def openDB(self, info=True):
		"""Open the database and creates the self.conn (connection)
		and self.curs (cursor) objects.

		Parameters:
			info (boolean, default True): show some output when opening DB

		Output:
			True
		"""
		if info:
			self.logger.info("Opening database: %s"%self.dbname)
		else:
			self.logger.debug("Opening database: %s"%self.dbname)
		self.conn = sqlite3.connect(self.dbname, check_same_thread=False)
		self.conn.row_factory = sqlite3.Row
		self.curs = self.conn.cursor()
		self.loadSubClasses()
		return True

	def reOpenDB(self):
		"""Not defined at this stage. Present in subclass PhysBiblioDB"""
		pass

	def closeDB(self, info=True):
		"""Close the database.

		Parameters:
			info (boolean, default True): show some output when opening DB
		"""
		if info:
			self.logger.info("Closing database...")
		else:
			self.logger.debug("Closing database...")
		self.conn.close()
		return True

	def checkUncommitted(self):
		"""Check if there are uncommitted changes.

		Output:
			True/False
		"""
		#return self.conn.in_transaction() #works only with sqlite > 3.2...
		return self.dbChanged

	def commit(self, verbose=True):
		"""Commit the changes.

		Parameters:
			verbose (boolean, default True):
				show some output when opening DB

		Output:
			True if successfull, False if an exception occurred
		"""
		try:
			self.conn.commit()
			self.dbChanged = False
			if verbose:
				self.logger.info("Database saved.")
			return True
		except Exception:
			self.logger.exception("Impossible to commit!")
			return False

	def undo(self, verbose=True):
		"""Undo the uncommitted changes
		and roll back to the last commit.

		Parameters:
			verbose (boolean, default True):
				show some output when opening DB

		Output:
			True if successfull, False if an exception occurred
		"""
		try:
			self.conn.rollback()
			self.dbChanged = False
			if verbose:
				self.logger.info("Rolled back to last commit.")
			return True
		except Exception:
			self.logger.exception("Impossible to rollback!")
			return False

	def connExec(self, query, data=None):
		"""Execute connection.

		Parameters:
			query (string): the query to be executed
			data (dictionary or list):
				the values of the parameters in the query

		Output:
			True if successfull, False if an exception occurred
		"""
		try:
			if data:
				self.conn.execute(query,data)
			else:
				self.conn.execute(query)
		except (OperationalError, ProgrammingError,
				DatabaseError, InterfaceError) as err:
			self.logger.exception('Connection error: %s\nquery: %s'%(
				err, query))
			return False
		except IntegrityError as err:
			self.logger.exception(
				'Cannot insert/update: ID exists!\n%s\nquery: %s'%(err, query))
			return False
		else:
			self.dbChanged = True
			return True

	def cursExec(self, query, data=None):
		"""Execute cursor.

		Parameters:
			query (string): the query to be executed
			data (dictionary or list):
				the values of the parameters in the query

		Output:
			True if successfull, False if an exception occurred
		"""
		try:
			if data:
				self.curs.execute(query, data)
			else:
				self.curs.execute(query)
		except Exception as err:
			self.logger.exception('Cursor error: %s\n'%err
				+ 'The query was: "%s"\n'%query
				+ ' and the parameters: %s'%(data,))
			return False
		else:
			return True

	def cursor(self):
		"""Function wrapper that returns the default cursor"""
		return self.curs

	def loadSubClasses(self):
		"""Not defined at this stage. Present in subclass PhysBiblioDB"""
		pass

	def checkExistingTables(self,
			wantedTables=["categories", "entries",
				"entryCats", "entryExps", "expCats",
				"experiments", "settings"]):
		"""Check that all the required tables
		are present in the database
		"""
		self.cursExec("SELECT name FROM sqlite_master WHERE type='table';")
		tables = [name[0] for name in self.curs]
		return not all(t in tables for t in wantedTables)

	def createTables(self, fieldsDict=None):
		"""Create tables for the database
		(and insert the default categories), if it is missing.

		Parameters:
			fieldsDict (default None):
				the structure of the tables (see physbiblio.tablesDef)
		"""
		if fieldsDict is None:
			fieldsDict = self.tableFields
		self.cursExec("SELECT name FROM sqlite_master WHERE type='table';")
		existingTables = [name[0] for name in self.curs]
		for q in fieldsDict.keys():
			if q in existingTables:
				continue
			command = "CREATE TABLE %s (\n"%q
			first = True
			for el in fieldsDict[q]:
				if first:
					first = False
				else:
					command += ",\n"
				command += " ".join(el)
			command += ");"
			self.logger.info(command + "\n")
			if not self.connExec(command):
				self.logger.critical("Create %s failed"%q)
		self.cursExec("select * from categories where "
			+ "idCat = 0 or idCat = 1\n")
		if len(self.curs.fetchall()) < 2:
			command = 'INSERT into categories ' \
				+ '(idCat, name, description, parentCat, ord) values ' \
				+ '(0,"Main","This is the main category. ' \
				+ 'All the other ones are subcategories of this one",0,0), ' \
				+ '(1,"Tags","Use this category to store tags ' \
				+ '(such as: ongoing projects, temporary cats,...)",0,0)\n'
			self.logger.info(command + "\n")
			if not self.connExec(command):
				self.logger.error("Insert main categories failed")
		self.commit()

	def checkCols(self):
		"""Run when new columns are added to the database with respect
		to previous versions of the software
		Check if bibdict column is present in entries table
		"""
		entryFields = self.cursExec("PRAGMA table_info(entries);")
		entriesCols = [name[1] for name in self.curs]
		if "bibdict" not in entriesCols:
			if self.connExec(
					"ALTER TABLE entries ADD COLUMN bibdict text;"):
				self.logger.info("New column in table 'entries': "
					+ "'bibdict' (text).")
				self.commit()
			else:
				pBLogger.error("Cannot alter table 'entries'!")
				self.undo()

class PhysBiblioDBSub():
	"""Uses PhysBiblioDB instance 'self.mainDB = parent'
	to act on the database.
	All the subcategories of PhysBiblioDB are defined
	starting from this one.
	"""

	def __init__(self, parent):
		"""Initialize DB class, connecting to
		the main PhysBiblioDB instance (parent).
		"""
		self.mainDB = parent
		#structure of the tables
		self.tableFields = self.mainDB.tableFields
		#names of the columns
		self.tableCols = {}
		for q in self.tableFields.keys():
			self.tableCols[q] = [ a[0] for a in self.tableFields[q] ]

		self.conn = self.mainDB.conn
		self.curs = self.mainDB.curs
		self.dbname = self.mainDB.dbname

		self.lastFetched = None
		self.catsHier = None

	def literal_eval(self, string):
		"""Wrapper for ast.literal_eval

		Parameters:
			string: the string to evaluate

		Output:
			a string or None
		"""
		try:
			if "[" in string and "]" in string:
				return ast.literal_eval(string.strip())
			elif "," in string:
				return ast.literal_eval("[%s]"%string.strip())
			else:
				return string.strip()
		except SyntaxError:
			self.mainDB.logger.warning(
				"Error in literal_eval with string '%s'"%string)
			return None

	def closeDB(self):
		"""Close the database (using PhysBiblioDB.close)"""
		self.mainDB.closeDB()

	def commit(self):
		"""Commit the changes (using PhysBiblioDB.commit)"""
		self.mainDB.commit()

	def connExec(self,query,data=None):
		"""Execute connection (see PhysBiblioDB.connExec)"""
		return self.mainDB.connExec(query, data=data)

	def cursExec(self, query, data=None):
		"""Execute cursor (see PhysBiblioDB.cursExec)"""
		return self.mainDB.cursExec(query, data=data)

	def cursor(self):
		"""Return the cursor"""
		return self.mainDB.cursor()



#!/usr/bin/env python
"""Test file for the physbiblio.pdf module.

This file is part of the physbiblio package.
"""
import sys
import traceback
import os
import shutil

if sys.version_info[0] < 3:
	import unittest2 as unittest
	from mock import patch, call
else:
	import unittest
	from unittest.mock import patch, call

try:
	from physbiblio.setuptests import *
	from physbiblio.config import pbConfig
	from physbiblio.database import pBDB, PhysBiblioDB
	from physbiblio.pdf import pBPDF
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
	print(traceback.format_exc())


pBPDF.pdfDir = os.path.join(pbConfig.dataPath, "testpdf_%s"%today_ymd)


@unittest.skipIf(skipTestsSettings.long, "Long tests")
class TestPdfMethods(unittest.TestCase):
	"""Test the methods and functions in the pdf module"""

	def test_fnames(self):
		"""Test names of folders and directories"""
		self.assertEqual(pBPDF.badFName(r'a\b/c:d*e?f"g<h>i|' + "j'"),
			"a_b_c_d_e_f_g_h_i_j_")
		self.assertEqual(pBPDF.badFName(True), "")
		self.assertEqual(pBPDF.getFileDir(r'a\b/c:d*e?f"g<h>i|' + "j'"),
			os.path.join(pBPDF.pdfDir, "a_b_c_d_e_f_g_h_i_j_"))
		testPaperFolder = pBPDF.getFileDir("abc.def")
		with patch('physbiblio.database.Entries.getField',
				return_value="12345678") as _mock:
			self.assertEqual(pBPDF.getFilePath("abc.def", "arxiv"),
				os.path.join(testPaperFolder, "12345678.pdf"))

	def test_manageFiles(self):
		"""Test creation, copy and deletion of files and folders"""
		emptyPdfName = os.path.join(pBPDF.pdfDir, "tests_%s.pdf"%today_ymd)
		pBPDF.createFolder("abc.def")
		self.assertTrue(os.path.exists(pBPDF.getFileDir("abc.def")))
		pBPDF.renameFolder("abc.def", "abc.fed")
		self.assertFalse(os.path.exists(pBPDF.getFileDir("abc.def")))
		self.assertTrue(os.path.exists(pBPDF.getFileDir("abc.fed")))
		open(emptyPdfName, 'a').close()
		with patch('physbiblio.database.Entries.getField',
				return_value="12345678") as _mock:
			pBPDF.copyNewFile("abc.fed", emptyPdfName, "arxiv")
			pBPDF.copyNewFile("abc.fed", emptyPdfName,
				customName="empty.pdf")
			self.assertTrue(os.path.exists(
				pBPDF.getFilePath("abc.fed", "arxiv")))
			os.remove(emptyPdfName)
			self.assertFalse(os.path.exists(emptyPdfName))
			pBPDF.copyToDir(pBPDF.pdfDir, "abc.fed", "arxiv")
			self.assertTrue(os.path.exists(
				os.path.join(pBPDF.pdfDir, "12345678.pdf")))
			pBPDF.mergePDFFolders("abc.fed", "cba.fed")
			self.assertTrue(os.path.exists(
				pBPDF.getFilePath("cba.fed", "arxiv")))
			self.assertTrue(pBPDF.removeFile("abc.fed", "arxiv"))
			self.assertFalse(pBPDF.removeFile("abc.fed", "arxiv"))
			os.remove(os.path.join(pBPDF.pdfDir, "12345678.pdf"))
			shutil.rmtree(pBPDF.getFileDir("abc.fed"))
			shutil.rmtree(pBPDF.getFileDir("cba.fed"))

	@unittest.skipIf(skipTestsSettings.online, "Online tests")
	def test_download(self):
		"""Test downloadArxiv"""
		with patch('physbiblio.database.Entries.getField',
				return_value="1806.11344") as _mock:
			self.assertTrue(pBPDF.downloadArxiv("abc.def"))
			self.assertTrue(pBPDF.checkFile("abc.def", "arxiv"))
			self.assertEqual(pBPDF.getExisting("abc.def", fullPath=False),
				["1806.11344.pdf"])
			self.assertEqual(pBPDF.getExisting("abc.def", fullPath=True),
				[os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344.pdf")])
			open(os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344"),
				"w").close()
			self.assertEqual(pBPDF.getExisting("abc.def", fullPath=False),
				["1806.11344.pdf"])
			self.assertTrue(pBPDF.removeFile("abc.def", "arxiv"))
			self.assertTrue(pBPDF.removeFile("abc.def", "file",
				os.path.join(pBPDF.getFileDir("abc.def"), "1806.11344")))
			self.assertFalse(pBPDF.checkFile("abc.def", "arxiv"))
		with patch('physbiblio.database.Entries.getField',
				side_effect=["1801.15000", "1801.15000", "", "", None, None]
				) as _mock:
			self.assertFalse(pBPDF.downloadArxiv("abc.def"))
			self.assertFalse(pBPDF.downloadArxiv("abc.def"))
			self.assertFalse(pBPDF.downloadArxiv("abc.def"))
		shutil.rmtree(pBPDF.getFileDir("abc.def"))

	def test_removeSpare(self):
		"""Test finding spare folders"""
		with patch('physbiblio.database.Entries.fetchCursor',
				return_value=[{"bibkey":"abc"}, {"bibkey":"def"}]) as _mock:
			for q in ["abc", "def", "ghi"]:
				pBPDF.createFolder(q)
				self.assertTrue(os.path.exists(pBPDF.getFileDir(q)))
			pBPDF.removeSparePDFFolders()
			for q in ["abc", "def"]:
				self.assertTrue(os.path.exists(pBPDF.getFileDir(q)))
			self.assertFalse(os.path.exists(pBPDF.getFileDir("ghi")))
		shutil.rmtree(pBPDF.pdfDir)

	def test_dirSize(self):
		"""test dirSize"""
		if sys.version_info[0] < 3:
			with patch("os.path.getsize",
					side_effect=[1, 100, 1, 100, 100]) as _gs,\
					patch("os.listdir",
						side_effect=[["a", "b"], ["c", "d", "e"]]) as _ld,\
					patch("os.path.isfile",
						side_effect=[True, False, True, True, False]) as _if,\
					patch("os.path.isdir",
						side_effect=[True, False]) as _id:
				self.assertEqual(pBPDF.dirSize("f"), 302)
				_gs.assert_has_calls([
					call('f'),
					call('f/a'),
					call('f/b'),
					call('f/b/c'),
					call('f/b/d')])
				_ld.assert_has_calls([call('f'), call('f/b')])
				_if.assert_has_calls([
					call('f/a'), call('f/b'), call('f/b/c'),
					call('f/b/d'), call('f/b/e')])
				_id.assert_has_calls([call('f/b'), call('f/b/e')])
			with patch("os.path.getsize",
						return_value=100) as _gs,\
					patch("os.listdir",
						side_effect=[["a", "b"], ["c", "d", "e"]]) as _ld,\
					patch("os.path.isfile",
						side_effect=[True, False, True, True, False]) as _if,\
					patch("os.path.isdir",
						side_effect=[True, False]) as _id:
				self.assertEqual(pBPDF.dirSize("f", dirs=False), 300)
				_gs.assert_has_calls([
					call('f/a'),
					call('f/b/c'),
					call('f/b/d')])
				_ld.assert_has_calls([call('f'), call('f/b')])
				_if.assert_has_calls([
					call('f/a'), call('f/b'), call('f/b/c'),
					call('f/b/d'), call('f/b/e')])
				_id.assert_has_calls([call('f/b'), call('f/b/e')])
		else:
			with patch("os.path.getsize",
					side_effect=[1, 1, 100, 100, 100]) as _gs,\
					patch("os.walk",
						return_value=[["f", ["b"],
							["a", "b/c", "b/d", "b/e"]]]) as _wa,\
					patch("os.path.isfile",
						side_effect=[True, True, True, False]) as _if:
				self.assertEqual(pBPDF.dirSize("f"), 302)
				_gs.assert_has_calls([
					call('f'),
					call('f/b'),
					call('f/a'),
					call('f/b/c'),
					call('f/b/d')])
				_wa.assert_called_once_with('f')
				_if.assert_has_calls([
					call('f/a'), call('f/b/c'),
					call('f/b/d'), call('f/b/e')])
			with patch("os.path.getsize",
					side_effect=[100, 100, 100]) as _gs,\
					patch("os.walk",
						return_value=[["f", ["b"],
							["a", "b/c", "b/d", "b/e"]]]) as _wa,\
					patch("os.path.isfile",
						side_effect=[True, True, True, False]) as _if:
				self.assertEqual(pBPDF.dirSize("f", dirs=False), 300)
				_gs.assert_has_calls([
					call('f/a'),
					call('f/b/c'),
					call('f/b/d')])
				_wa.assert_called_once_with('f')
				_if.assert_has_calls([
					call('f/a'), call('f/b/c'),
					call('f/b/d'), call('f/b/e')])

	def test_getSizeWUnits(self):
		"""test getSizeWUnits"""
		with patch("logging.Logger.warning") as _w:
			self.assertEqual(pBPDF.getSizeWUnits(2048**2), "4.00MB")
			self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="mb"),
				"4.00MB")
			self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="kb"),
				"4096.00KB")
			self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="PB"),
				"0.00PB")
			self.assertEqual(
				pBPDF.getSizeWUnits(2048**2, units="gb", fmt="%.3f"),
				"0.004GB")
			_w.assert_not_called()
		with patch("logging.Logger.warning") as _w:
			self.assertEqual(pBPDF.getSizeWUnits(2048**2, units="m"),
				"4.00MB")
			_w.assert_called_once_with("Invalid units. Changing to 'MB'.")
		with patch("logging.Logger.warning") as _w:
			self.assertEqual(pBPDF.getSizeWUnits("a"), "nan")
			_w.assert_called_once_with('Invalid size. It must be a number!')
		with patch("logging.Logger.warning") as _w:
			self.assertEqual(pBPDF.getSizeWUnits(2048**2, fmt="%"), "4.00MB")
			_w.assert_called_once_with("Invalid format. Using '%.2f'")


if __name__=='__main__':
	unittest.main()



"""Module with the classes and functions that manage
the categories windows and panels.

This file is part of the physbiblio package.
"""
import traceback
from PySide2.QtCore import Qt, QTimer
from PySide2.QtGui import QCursor
from PySide2.QtWidgets import \
	QAction, QDialog, QLineEdit, QPushButton, QTreeView, \
	QVBoxLayout, QToolTip

try:
	from physbiblio.errors import pBLogger
	from physbiblio.database import pBDB, cats_alphabetical
	from physbiblio.config import pbConfig
	from physbiblio.view import pBView
	from physbiblio.gui.errorManager import pBGUILogger
	from physbiblio.gui.basicDialogs import \
		askYesNo
	from physbiblio.gui.commonClasses import \
		EditObjectWindow, LeafFilterProxyModel, PBLabel, PBMenu, \
		NamedElement, NamedNode, TreeModel
	import physbiblio.gui.resourcesPyside2
except ImportError:
	print("Could not find physbiblio and its modules!")
	print(traceback.format_exc())


def editCategory(parentObject,
		mainWinObject,
		editIdCat=None,
		useParentCat=None):
	"""Open a dialog (`EditCategoryDialog`) to edit a category
	and process the output.

	Parameters:
		parentObject: the parent widget
		mainWinObject: the object which has
			the `statusBarMessage` and `setWindowTitle` methods
		editIdCat: the id of the category to be edited,
			or `None` to create a new category
		useParentCat: the parent category (if any)
			of the one to be edited
	"""
	if editIdCat is not None:
		edit = pBDB.cats.getDictByID(editIdCat)
	else:
		edit = None
	newCatWin = EditCategoryDialog(parentObject,
		category=edit,
		useParentCat=useParentCat)
	newCatWin.exec_()
	if newCatWin.result:
		data = {}
		for k, v in newCatWin.textValues.items():
			if k == "parentCat":
				try:
					s = str(newCatWin.selectedCats[0])
				except IndexError:
					s = "0"
			else:
				s = "%s"%v.text()
			data[k] = s
		if data["name"].strip() != "":
			if "idCat" in data.keys():
				pBLogger.info("Updating category %s..."%data["idCat"])
				pBDB.cats.update(data, data["idCat"])
			else:
				pBDB.cats.insert(data)
			message = "Category saved"
			mainWinObject.setWindowTitle("PhysBiblio*")
			try:
				parentObject.recreateTable()
			except AttributeError:
				pBLogger.debug("parentObject has no attribute 'recreateTable'",
					exc_info=True)
		else:
			message = "ERROR: empty category name"
	else:
		message = "No modifications to categories"
	try:
		mainWinObject.statusBarMessage(message)
	except AttributeError:
		pBLogger.debug("mainWinObject has no attribute 'statusBarMessage'",
			exc_info=True)


def deleteCategory(parentObject, mainWinObject, idCat, name):
	"""Ask confirmation and eventually delete the selected category

	Parameters:
		parentObject: the parent widget
		mainWinObject: the object which has
			the `statusBarMessage` and `setWindowTitle` methods
		idCat: the id of the category to be deleted
		name: the name of the category to be deleted
	"""
	if askYesNo("Do you really want to delete this category "
			+ "(ID = '%s', name = '%s')?"%(idCat, name)):
		pBDB.cats.delete(int(idCat))
		mainWinObject.setWindowTitle("PhysBiblio*")
		message = "Category deleted"
		try:
			parentObject.recreateTable()
		except AttributeError:
			pBLogger.debug("parentObject has no attribute 'recreateTable'",
				exc_info=True)
	else:
		message = "Nothing changed"
	try:
		mainWinObject.statusBarMessage(message)
	except AttributeError:
		pBLogger.debug("mainWinObject has no attribute 'statusBarMessage'",
			exc_info=True)


class CatsModel(TreeModel):
	"""Model for the categories tree"""

	def __init__(self,
			cats,
			rootElements,
			parent=None,
			previous=[],
			multipleRecords=False):
		"""Initialize the model, save the data and the initial selection

		Parameters:
			cats: the list of categories records from the database
			rootElements: the list of root elements of the tree
			parent: the parent widget
				(default: None)
			previous: the list of previously selected categories
				(default: an empty list)
			multipleRecords: activate tristate for selecting categories
				for multiple records
				(default False)
		"""
		self.cats = cats
		self.rootElements = rootElements
		TreeModel.__init__(self)
		self.parentObj = parent
		self.selectedCats = {}
		self.previousSaved = {}
		for cat in self.cats:
			self.selectedCats[cat["idCat"]] = False
			self.previousSaved[cat["idCat"]] = False
		for prevIx in previous:
			if prevIx in [cat["idCat"] for cat in self.cats]:
				if multipleRecords:
					self.previousSaved[prevIx] = True
					self.selectedCats[prevIx] = "p"
				else:
					self.selectedCats[prevIx] = True
			else:
				pBLogger.warning("Invalid idCat in previous selection: "
					+ "%s"%prevIx)

	def _getRootNodes(self):
		"""Obtain the list of named nodes which represent the root
		elements of the tree

		Output:
			a list of `NamedNode`s
		"""
		return [NamedNode(elem, None, index)
			for index, elem in enumerate(self.rootElements)]

	def columnCount(self, parent):
		"""Return the number of columns, fixed to one"""
		return 1

	def data(self, index, role):
		"""Return the cell data for the given index and role

		Parameters:
			index: the `QModelIndex` for which the data are required
			role: the desired Qt role for the data

		Output:
			None if the index or the role are not valid,
			the cell content of properties otherwise
		"""
		if not index.isValid():
			return None
		row = index.row()
		column = index.column()
		value = index.internalPointer().element.text
		idCat = index.internalPointer().element.idCat
		if role == Qt.CheckStateRole and \
				column == 0 and \
				hasattr(self.parentObj, "askCats") and \
				self.parentObj.askCats:
			if self.previousSaved[idCat] == True and \
					self.selectedCats[idCat] == "p":
				return Qt.PartiallyChecked
			elif self.selectedCats[idCat] == False:
				return Qt.Unchecked
			else:
				return Qt.Checked
		if role == Qt.EditRole:
			return value
		if role == Qt.DisplayRole:
			return value
		return None

	def flags(self, index):
		"""Return the cell flags for the given index

		Parameter:
			index: the `QModelIndex` for which the data are required

		Output:
			None if the index or the role are not valid,
			the cell flags otherwise
		"""
		if not index.isValid():
			return None
		if index.column() == 0 and (hasattr(self.parentObj, "askCats")
				and self.parentObj.askCats):
			return Qt.ItemIsUserCheckable | Qt.ItemIsEditable | \
				Qt.ItemIsEnabled | Qt.ItemIsSelectable
		else:
			return Qt.ItemIsEnabled | Qt.ItemIsSelectable

	def headerData(self, section, orientation, role):
		"""Return the header data given section, orientation and role

		Parameters:
			section: the section number
			orientation: horizontal or vertical header
			role: the requested role

		Output:
			the column name ("Name", if the parameters correspond)
			or None
		"""
		if orientation == Qt.Horizontal and role == Qt.DisplayRole \
			and section == 0:
			return 'Name'
		return None

	def setData(self, index, value, role):
		"""Set the cell data for the given index and role

		Parameters:
			index: the `QModelIndex` for which the data are required
			value: the new data value
			role: the desired Qt role for the data

		Output:
			True if correctly completed,
			False if the `index` is not valid
		"""
		if not index.isValid():
			return False
		idCat = index.internalPointer().element.idCat
		if role == Qt.CheckStateRole and index.column() == 0:
			self.previousSaved[idCat] = False
			if value == Qt.Checked:
				self.selectedCats[idCat] = True
			else:
				self.selectedCats[idCat] = False
		self.dataChanged.emit(index, index)
		return True


class CatsTreeWindow(QDialog):
	"""Extension of `QDialog` that shows the categories tree"""

	def __init__(self,
			parent=None,
			askCats=False,
			askForBib=None,
			askForExp=None,
			expButton=True,
			previous=[],
			single=False,
			multipleRecords=False):
		"""Initialize instance parameters and call the function that
		creates the layout.

		Parameters:
			parent (default None): the parent widget
			askCats (default False): if True, enable checkboxes for
				selection of categories
			askForBib (default None): the optional key which identifies
				in the database the bibtex entry for which categories
				are being selected
			askForExp (default None): the optional ID which identifies
				in the database the experiment for which categories
				are being selected
			expButton (default True): if True, add a button to accept
				the widget content and later ask for experiments
			previous (default []): the list of categories that
				must be selected at the beginning
			single (default False): if True, only allow the selection
				of a single category (the parent category, typically).
				Multiple checkboxes can be selected,
				but only the first one will be considered
			multipleRecords: used when dealing with categories
				corresponding to multiple records. Activate a tristate
				checkbox for the initial list of categories, which are
				typically not the same for all the elements in the list
		"""
		QDialog.__init__(self, parent)
		self.setWindowTitle("Categories")
		self.currLayout = QVBoxLayout(self)
		self.setLayout(self.currLayout)
		self.askCats = askCats
		self.askForBib = askForBib
		self.askForExp = askForExp
		self.expButton = expButton
		self.previous = previous
		self.single = single
		self.multipleRecords = multipleRecords

		self.setMinimumWidth(400)
		self.setMinimumHeight(600)

		self.createForm()

	def populateAskCat(self):
		"""If selection of categories is allowed, add some information
		on the bibtex/experiment for which the categories are requested
		and a simple message, then create few required empty lists
		"""
		if self.askCats:
			if self.askForBib is not None:
				try:
					bibitem = pBDB.bibs.getByBibkey(self.askForBib,
						saveQuery=False)[0]
				except IndexError:
					pBGUILogger.warning("The entry '%s' "%self.askForBib
						+ "is not in the database!", exc_info=True)
					return
				try:
					if bibitem["inspire"] != "" and \
							bibitem["inspire"] is not None:
						link = "<a href='%s'>%s</a>"%(
							pBView.getLink(self.askForBib, "inspire"),
							self.askForBib)
					elif bibitem["arxiv"] != "" and \
							bibitem["arxiv"] is not None:
						link = "<a href='%s'>%s</a>"%(
							pBView.getLink(self.askForBib, "arxiv"),
							self.askForBib)
					elif bibitem["doi"] != "" and \
							bibitem["doi"] is not None:
						link = "<a href='%s'>%s</a>"%(
							pBView.getLink(self.askForBib, "doi"),
							self.askForBib)
					else:
						link = self.askForBib
					bibtext = PBLabel("Mark categories for the following "
						+ "entry:<br><b>key</b>:<br>%s<br>"%link
						+ "<b>author(s)</b>:<br>%s<br>"%bibitem["author"]
						+ "<b>title</b>:<br>%s<br>"%bibitem["title"])
				except KeyError:
					bibtext = PBLabel("Mark categories for the following "
						+ "entry:<br><b>key</b>:<br>%s<br>"%(self.askForBib))
				self.currLayout.addWidget(bibtext)
			elif self.askForExp is not None:
				try:
					expitem = pBDB.exps.getByID(self.askForExp)[0]
				except IndexError:
					pBGUILogger.warning("The experiment ID %s"%self.askForExp
						+ " is not in the database!", exc_info=True)
					return
				try:
					exptext = PBLabel("Mark categories for the following "
						+ "experiment:<br><b>id</b>:<br>%s<br>"%self.askForExp
						+ "<b>name</b>:<br>%s<br>"%expitem["name"]
						+ "<b>comments</b>:<br>%s<br>"%expitem["comments"])
				except KeyError:
					exptext = PBLabel("Mark categories for the following "
						+ "experiment:<br><b>id</b>:<br>%s<br>"%(
							self.askForExp))
				self.currLayout.addWidget(exptext)
			else:
				if self.single:
					comment = PBLabel("Select the desired category "
						+ "(only the first one will be considered):")
				else:
					comment = PBLabel("Select the desired categories:")
				self.currLayout.addWidget(comment)
			self.marked = []
			self.parent().selectedCats = []
		return True

	def onCancel(self):
		"""Reject the dialog content and close the window"""
		self.result	= False
		self.close()

	def onOk(self, exps=False):
		"""Accept the dialog content
		(update the list of selected categories)
		and close the window.
		May set `self.result` to "Exps"
		for later opening of a new dialog to ask for experiments.

		Parameter:
			exps (default False): if True, set the result to "Exps",
				otherwise to "Ok"
		"""
		self.parent().selectedCats = [idC \
			for idC in self.root_model.selectedCats.keys() \
			if self.root_model.selectedCats[idC] == True]
		self.parent().previousUnchanged = [idC \
			for idC in self.root_model.previousSaved.keys() \
			if self.root_model.previousSaved[idC] == True]

		if self.single and \
				len(self.parent().selectedCats) > 1 and \
				self.parent().selectedCats[0] == 0:
			self.parent().selectedCats.pop(0)
			self.parent().selectedCats = [self.parent().selectedCats[0]]
		if exps:
			self.result	= "Exps"
		else:
			self.result	= "Ok"
		self.close()

	def changeFilter(self, string):
		"""When the filter `QLineEdit` is changed,
		update the `LeafFilterProxyModel` regexp filter

		Parameter:
			string: the new filter string
		"""
		self.proxyModel.setFilterRegExp(str(string))
		self.tree.expandAll()

	def onAskExps(self):
		"""Action to perform when the selection of categories
		will be folloed by the selection of experiments.
		Call `self.onOk` with `exps = True`.
		"""
		self.onOk(exps = True)

	def onNewCat(self):
		"""Action to perform when the creation
		of a new category is requested
		"""
		editCategory(self, self.parent())

	def keyPressEvent(self, e):
		"""Manage the key press events.
		Do nothing unless `Esc` is pressed:
		in this case close the dialog
		"""
		if e.key() == Qt.Key_Escape:
			self.close()

	def createForm(self):
		"""Create the dialog content, connect the model to the view
		and eventually add the buttons at the end
		"""
		self.populateAskCat()

		catsTree = pBDB.cats.getHier()

		self.filterInput = QLineEdit("",  self)
		self.filterInput.setPlaceholderText("Filter categories")
		self.filterInput.textChanged.connect(self.changeFilter)
		self.currLayout.addWidget(self.filterInput)
		self.filterInput.setFocus()

		self.tree = QTreeView(self)
		self.currLayout.addWidget(self.tree)
		self.tree.setMouseTracking(True)
		self.tree.entered.connect(self.handleItemEntered)

		catsNamedTree = self._populateTree(catsTree[0], 0)

		self.root_model = CatsModel(pBDB.cats.getAll(),
			[catsNamedTree],
			self,
			self.previous,
			multipleRecords=self.multipleRecords)
		self.proxyModel = LeafFilterProxyModel(self)
		self.proxyModel.setSourceModel(self.root_model)
		self.proxyModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
		self.proxyModel.setSortCaseSensitivity(Qt.CaseInsensitive)
		self.proxyModel.setFilterKeyColumn(-1)
		self.tree.setModel(self.proxyModel)

		self.tree.expandAll()

		self.tree.setHeaderHidden(True)
		#self.tree.doubleClicked.connect(self.askAndPerformAction)

		self.newCatButton = QPushButton('Add new category', self)
		self.newCatButton.clicked.connect(self.onNewCat)
		self.currLayout.addWidget(self.newCatButton)

		if self.askCats:
			self.acceptButton = QPushButton('OK', self)
			self.acceptButton.clicked.connect(self.onOk)
			self.currLayout.addWidget(self.acceptButton)

			if self.expButton:
				self.expsButton = QPushButton('Ask experiments', self)
				self.expsButton.clicked.connect(self.onAskExps)
				self.currLayout.addWidget(self.expsButton)

			# cancel button
			self.cancelButton = QPushButton('Cancel', self)
			self.cancelButton.clicked.connect(self.onCancel)
			self.cancelButton.setAutoDefault(True)
			self.currLayout.addWidget(self.cancelButton)

	def _populateTree(self, children, idCat):
		"""Read the list of categories recursively and
		populate the categories tree

		Parameters:
			children: the list of children categories
				of the currently considered one
			idCat: the id of the current category
		"""
		name = pBDB.cats.getByID(idCat)[0]["name"]
		children_list = []
		for child in cats_alphabetical(children, pBDB):
			child_item = self._populateTree(children[child], child)
			children_list.append(child_item)
		return NamedElement(idCat, name, children_list)

	def handleItemEntered(self, index):
		"""Process event when mouse enters an item and
		create a `QTooltip` which describes the category, with a timer

		Parameter:
			index: a `QModelIndex` instance
		"""
		if index.isValid():
			row = index.row()
		else:
			return
		idCat, catName = self.proxyModel.sibling(row, 0, index) \
			.data().split(": ")
		idCat = idCat.strip()
		try:
			self.timer.stop()
			QToolTip.showText(QCursor.pos(), "", self.tree.viewport())
		except AttributeError:
			pass
		try:
			catData = pBDB.cats.getByID(idCat)[0]
		except IndexError:
			pBGUILogger.exception("Failed in finding category")
			return
		self.timer = QTimer(self)
		self.timer.setSingleShot(True)
		self.timer.timeout.connect(lambda: QToolTip.showText(
			QCursor.pos(),
			"{idC}: {cat}\nCorresponding entries: ".format(
				idC = idCat,
				cat = catData["name"])
			+ "{en}\nAssociated experiments: {ex}".format(
				en = pBDB.catBib.countByCat(idCat),
				ex = pBDB.catExp.countByCat(idCat)),
			self.tree.viewport(),
			self.tree.visualRect(index),
			3000))
		self.timer.start(500)

	def contextMenuEvent(self, event):
		"""Create a right click menu with few actions
		on the selected category

		Parameter:
			event: a `QEvent`
		"""
		index = self.tree.selectedIndexes()[0]
		if index.isValid():
			row = index.row()
		else:
			return
		idCat, catName = self.proxyModel.sibling(row, 0, index) \
			.data().split(": ")
		idCat = idCat.strip()

		menu = PBMenu()
		self.menu = menu
		titAction = QAction("--Category: %s--"%catName)
		titAction.setDisabled(True)
		bibAction = QAction("Open list of corresponding entries")
		modAction = QAction("Modify")
		delAction = QAction("Delete")
		subAction = QAction("Add subcategory")
		menu.possibleActions = [
			titAction, None,
			bibAction, None,
			modAction, delAction, None,
			subAction
			]
		menu.fillMenu()

		action = menu.exec_(event.globalPos())

		if action == bibAction:
			searchDict = {"cats": {"id": [idCat], "operator": "and"}}
			self.parent().reloadMainContent(
				pBDB.bibs.fetchFromDict(searchDict).lastFetched)
		elif action == modAction:
			editCategory(self, self.parent(), idCat)
		elif action == delAction:
			deleteCategory(self, self.parent(), idCat, catName)
		elif action == subAction:
			editCategory(self, self.parent(), useParent=idCat)

	def cleanLayout(self):
		"""Delete the previous widgets"""
		while True:
			o = self.layout().takeAt(0)
			if o is None: break
			o.widget().deleteLater()

	def recreateTable(self):
		"""Delete the previous widgets and recreate them with new data"""
		self.cleanLayout()
		self.createForm()


class EditCategoryDialog(EditObjectWindow):
	"""Create a window for editing or creating a category"""

	def __init__(self, parent=None, category=None, useParentCat=None):
		"""Set some properties and create the form content

		Parameters:
			parent: the parent widget
			category: a category record from the database
			useParentCat: the id of the parent category
				of the given one in the tree
		"""
		super(EditCategoryDialog, self).__init__(parent)
		if category is None:
			self.data = {}
			for k in pBDB.tableCols["categories"]:
				self.data[k] = ""
		else:
			self.data = category
		if useParentCat is not None:
			self.data["parentCat"] = useParentCat
			self.selectedCats = [useParentCat]
		elif category is not None:
			try:
				self.selectedCats = [pBDB.cats.getParent(
					category["idCat"])[0][0]]
			except IndexError:
				self.selectedCats = [0]
		else:
			self.selectedCats = [0]
		self.createForm()

	def onAskParent(self):
		"""Open a `CatsTreeWindow` and process its output"""
		selectCats = CatsTreeWindow(parent=self,
			askCats=True,
			expButton=False,
			single=True,
			previous=self.selectedCats)
		selectCats.exec_()
		if selectCats.result == "Ok":
			try:
				val = self.selectedCats[0]
				self.textValues["parentCat"].setText(
					"%s - %s"%(str(val), pBDB.cats.getByID(val)[0]["name"]))
			except IndexError:
				self.textValues["parentCat"].setText("Select parent")

	def createForm(self):
		"""Prepare the window widgets"""
		self.setWindowTitle('Edit category')

		i = 0
		for k in pBDB.tableCols["categories"]:
			val = self.data[k]
			if k != "idCat" or (k == "idCat" and self.data[k] != ""):
				i += 1
				self.currGrid.addWidget(PBLabel(
					pBDB.descriptions["categories"][k]), i*2 - 1, 0)
				self.currGrid.addWidget(PBLabel("(%s)"%k), i*2 - 1, 1)
				if k == "parentCat":
					try:
						val = self.selectedCats[0]
						self.textValues[k] = QPushButton(
							"%s - %s"%(str(val),
								pBDB.cats.getByID(val)[0]["name"]), self)
					except IndexError:
						self.textValues[k] = QPushButton("Select parent", self)
					self.textValues[k].clicked.connect(self.onAskParent)
				else:
					self.textValues[k] = QLineEdit(str(val))
				if k == "idCat":
					self.textValues[k].setEnabled(False)
				self.currGrid.addWidget(self.textValues[k], i*2, 0, 1, 2)

		# OK button
		self.acceptButton = QPushButton('OK', self)
		self.acceptButton.clicked.connect(self.onOk)
		self.currGrid.addWidget(self.acceptButton, i*2 + 1, 1)

		# cancel button
		self.cancelButton = QPushButton('Cancel', self)
		self.cancelButton.clicked.connect(self.onCancel)
		self.cancelButton.setAutoDefault(True)
		self.currGrid.addWidget(self.cancelButton, i*2 + 1, 0)

		self.setGeometry(100, 100, 400, 50*i)
		self.centerWindow()
